import logging
import select
import socket
import threading
import time
from abc import abstractmethod
from typing import Optional

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import ConnectionFailedError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider


class TcpForwarderBase:
    """
    Allows forwarding local tcp connection into the device via a given lockdown connection
    """

    MAX_FORWARDED_CONNECTIONS = 200
    TIMEOUT = 1

    def __init__(self, src_port: int, listening_event: Optional[threading.Event] = None):
        """
        Initialize a new tcp forwarder

        :param src_port: tcp port to listen on
        :param enable_ssl: enable ssl wrapping for the transferred data
        :param listening_event: event to fire when the listening occurred
        """
        self.logger = logging.getLogger(__name__)
        self.src_port = src_port
        self.server_socket = None
        self.inputs = []
        self.stopped = threading.Event()
        self.listening_event = listening_event

        # dictionaries containing the required maps to transfer data between each local
        # socket to its remote socket and vice versa
        self.connections = {}

    def start(self, address="127.0.0.1"):
        """forward each connection from given local machine port to remote device port"""
        # create local tcp server socket
        self.server_socket = socket.socket()
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((address, self.src_port))
        self.server_socket.listen(self.MAX_FORWARDED_CONNECTIONS)
        self.server_socket.setblocking(False)

        self.inputs = [self.server_socket]
        if self.listening_event:
            self.listening_event.set()

        try:
            while self.inputs:
                # will only perform the socket select on the inputs. the outputs will handled
                # as synchronous blocking
                readable, _writable, exceptional = select.select(self.inputs, [], self.inputs, self.TIMEOUT)
                if self.stopped.is_set():
                    self.logger.debug("Closing since stopped is set")
                    break

                closed_sockets = set()
                for current_sock in readable:
                    self.logger.debug("Processing %r", current_sock)
                    if current_sock is self.server_socket:
                        self._handle_server_connection()
                    else:
                        if current_sock not in closed_sockets:
                            try:
                                self._handle_data(current_sock, closed_sockets)
                            except ConnectionResetError:
                                self.logger.error("Error when handling data")
                                self._handle_close_or_error(current_sock)
                        else:
                            self.logger.debug("Is closed")

                for current_sock in exceptional:
                    self.logger.error("Sock failed: %r", current_sock)
                    self._handle_close_or_error(current_sock)
        finally:
            self.logger.info("Closing everything")
            # on stop, close all currently opened sockets
            for current_sock in self.inputs:
                current_sock.close()

    def _handle_close_or_error(self, from_sock):
        """if an error occurred its time to close the two sockets"""
        other_sock = self.connections[from_sock]

        other_sock.close()
        from_sock.close()
        self.inputs.remove(other_sock)
        self.inputs.remove(from_sock)

        self.logger.info(f"connection {other_sock} was closed")

    def _handle_data(self, from_sock, closed_sockets):
        self.logger.debug(f"Handling data from {from_sock}")
        try:
            data = from_sock.recv(1024)
            if not data:
                raise ConnectionResetError("Connection closed by the peer.")
        except BlockingIOError:
            self.logger.warning(f"Non-blocking read failed on {from_sock}, retrying later.")
            return
        except OSError:
            self.logger.error(f"Error reading from socket {from_sock}")
            self._handle_close_or_error(from_sock)
            closed_sockets.add(from_sock)
            return

        other_sock = self.connections.get(from_sock)
        if not other_sock:
            self.logger.error(f"No connection mapping found for {from_sock}.")
            return

        try:
            total_sent = 0
            while total_sent < len(data):
                try:
                    sent = other_sock.send(data[total_sent:])
                    total_sent += sent
                except BlockingIOError:
                    self.logger.warning(f"Socket buffer full for {other_sock}, retrying in 100ms.")
                    time.sleep(0.1)  # Introduce a small delay
                except BrokenPipeError:
                    self.logger.error(f"Broken pipe error on {other_sock}.")
                    raise
        except OSError:
            self.logger.error("Unhandled error while forwarding data")
            self._handle_close_or_error(from_sock)
            closed_sockets.add(from_sock)
            closed_sockets.add(other_sock)

    @abstractmethod
    def _establish_remote_connection(self) -> socket.socket:
        pass

    def _handle_server_connection(self):
        """accept the connection from local machine and attempt to connect at remote"""
        local_connection, _client_address = self.server_socket.accept()
        local_connection.setblocking(False)

        try:
            remote_connection = self._establish_remote_connection()
        except ConnectionFailedError:
            self.logger.error(f"failed to connect to port: {self.dst_port}")
            local_connection.close()
            return

        remote_connection.setblocking(False)

        # append the newly created sockets into input list
        self.inputs.append(local_connection)
        self.inputs.append(remote_connection)

        # and store a map of which local connection is transferred to which remote one
        self.connections[remote_connection] = local_connection
        self.connections[local_connection] = remote_connection

        self.logger.info("connection established from local to remote")

    def stop(self):
        """stop forwarding"""
        self.stopped.set()


class UsbmuxTcpForwarder(TcpForwarderBase):
    """
    Allows forwarding local tcp connection into the device via a given lockdown connection
    """

    def __init__(
        self,
        serial: str,
        dst_port: int,
        src_port: int,
        listening_event: Optional[threading.Event] = None,
        usbmux_connection_type: Optional[str] = None,
        usbmux_address: Optional[str] = None,
    ):
        """
        Initialize a new tcp forwarder

        :param serial: device serial
        :param dst_port: tcp port to connect to each new connection via the supplied lockdown object
        :param src_port: tcp port to listen on
        :param listening_event: event to fire when the listening occurred
        :param usbmux_connection_type: preferred connection type
        :param usbmux_address: usbmuxd address
        """
        super().__init__(src_port, listening_event)
        self.serial = serial
        self.dst_port = dst_port
        self.usbmux_connection_type = usbmux_connection_type
        self.usbmux_address = usbmux_address

    def _establish_remote_connection(self) -> socket.socket:
        # connect directly using usbmuxd
        mux_device = usbmux.select_device(
            self.serial, connection_type=self.usbmux_connection_type, usbmux_address=self.usbmux_address
        )
        self.logger.debug("Selected device: %r", mux_device)
        if mux_device is None:
            raise ConnectionFailedError()
        return mux_device.connect(self.dst_port, usbmux_address=self.usbmux_address)


class LockdownTcpForwarder(TcpForwarderBase):
    """
    Allows forwarding local tcp connection into the device via a given lockdown connection
    """

    def __init__(
        self,
        service_provider: LockdownServiceProvider,
        src_port: int,
        service_name: str,
        listening_event: Optional[threading.Event] = None,
    ):
        """
        Initialize a new tcp forwarder

        :param src_port: tcp port to listen on
        :param service_name: service name to connect to
        :param listening_event: event to fire when the listening occurred
        """
        super().__init__(src_port, listening_event)
        self.service_provider = service_provider
        self.service_name = service_name

    def _establish_remote_connection(self) -> socket.socket:
        return self.service_provider.start_lockdown_developer_service(self.service_name).socket



import abc
import plistlib
import socket
import time
from dataclasses import dataclass
from typing import Optional

from construct import (
    Const,
    CString,
    Enum,
    FixedSized,
    GreedyBytes,
    Int16ul,
    Int32ul,
    Padding,
    Prefixed,
    StreamError,
    Struct,
    Switch,
    this,
)

from pymobiledevice3.exceptions import (
    BadCommandError,
    BadDevError,
    ConnectionFailedError,
    ConnectionFailedToUsbmuxdError,
    MuxException,
    MuxVersionError,
    NotPairedError,
)
from pymobiledevice3.osu.os_utils import get_os_utils

usbmuxd_version = Enum(
    Int32ul,
    BINARY=0,
    PLIST=1,
)

usbmuxd_result = Enum(
    Int32ul,
    OK=0,
    BADCOMMAND=1,
    BADDEV=2,
    CONNREFUSED=3,
    NOSUCHSERVICE=4,
    BADVERSION=6,
)

usbmuxd_msgtype = Enum(
    Int32ul,
    RESULT=1,
    CONNECT=2,
    LISTEN=3,
    ADD=4,
    REMOVE=5,
    PAIRED=6,
    PLIST=8,
)

usbmuxd_header = Struct(
    "version" / usbmuxd_version,  # protocol version
    "message" / usbmuxd_msgtype,  # message type
    "tag" / Int32ul,  # responses to this query will echo back this tag
)

usbmuxd_request = Prefixed(
    Int32ul,
    Struct(
        "header" / usbmuxd_header,
        "data"
        / Switch(
            this.header.message,
            {
                usbmuxd_msgtype.CONNECT: Struct(
                    "device_id" / Int32ul,
                    "port" / Int16ul,  # TCP port number
                    "reserved" / Const(0, Int16ul),
                ),
                usbmuxd_msgtype.PLIST: GreedyBytes,
            },
        ),
    ),
    includelength=True,
)

usbmuxd_device_record = Struct(
    "device_id" / Int32ul,
    "product_id" / Int16ul,
    "serial_number" / FixedSized(256, CString("ascii")),
    Padding(2),
    "location" / Int32ul,
)

usbmuxd_response = Prefixed(
    Int32ul,
    Struct(
        "header" / usbmuxd_header,
        "data"
        / Switch(
            this.header.message,
            {
                usbmuxd_msgtype.RESULT: Struct(
                    "result" / usbmuxd_result,
                ),
                usbmuxd_msgtype.ADD: usbmuxd_device_record,
                usbmuxd_msgtype.REMOVE: Struct(
                    "device_id" / Int32ul,
                ),
                usbmuxd_msgtype.PLIST: GreedyBytes,
            },
        ),
    ),
    includelength=True,
)


@dataclass
class MuxDevice:
    devid: int
    serial: str
    connection_type: str

    def connect(self, port: int, usbmux_address: Optional[str] = None) -> socket.socket:
        mux = create_mux(usbmux_address=usbmux_address)
        try:
            return mux.connect(self, port)
        except:
            mux.close()
            raise

    @property
    def is_usb(self) -> bool:
        return self.connection_type == "USB"

    @property
    def is_network(self) -> bool:
        return self.connection_type == "Network"

    def matches_udid(self, udid: str) -> bool:
        return self.serial.replace("-", "") == udid.replace("-", "")


class SafeStreamSocket:
    """wrapper to native python socket object to be used with construct as a stream"""

    def __init__(self, address, family):
        self._offset = 0
        self.sock = socket.socket(family, socket.SOCK_STREAM)
        self.sock.connect(address)

    def send(self, msg: bytes) -> int:
        self._offset += len(msg)
        self.sock.sendall(msg)
        return len(msg)

    def recv(self, size: int) -> bytes:
        msg = b""
        while len(msg) < size:
            chunk = self.sock.recv(size - len(msg))
            self._offset += len(chunk)
            if not chunk:
                raise MuxException("socket connection broken")
            msg += chunk
        return msg

    def close(self) -> None:
        self.sock.close()

    def settimeout(self, interval: float) -> None:
        self.sock.settimeout(interval)

    def setblocking(self, blocking: bool) -> None:
        self.sock.setblocking(blocking)

    def tell(self) -> int:
        return self._offset

    read = recv
    write = send


class MuxConnection:
    # used on Windows
    ITUNES_HOST = ("127.0.0.1", 27015)

    # used for macOS and Linux
    USBMUXD_PIPE = "/var/run/usbmuxd"

    @staticmethod
    def create_usbmux_socket(usbmux_address: Optional[str] = None) -> SafeStreamSocket:
        try:
            if usbmux_address is not None:
                if ":" in usbmux_address:
                    # assume tcp address
                    hostname, port = usbmux_address.split(":")
                    port = int(port)
                    address = (hostname, port)
                    family = socket.AF_INET
                else:
                    # assume unix domain address
                    address = usbmux_address
                    family = socket.AF_UNIX
            else:
                address, family = get_os_utils().usbmux_address
            return SafeStreamSocket(address, family)
        except ConnectionRefusedError as e:
            raise ConnectionFailedToUsbmuxdError() from e

    @staticmethod
    def create(usbmux_address: Optional[str] = None):
        # first attempt to connect with possibly the wrong version header (plist protocol)
        sock = MuxConnection.create_usbmux_socket(usbmux_address=usbmux_address)

        try:
            message = usbmuxd_request.build({
                "header": {"version": usbmuxd_version.PLIST, "message": usbmuxd_msgtype.PLIST, "tag": 1},
                "data": plistlib.dumps({"MessageType": "ReadBUID"}),
            })
            sock.send(message)
            response = usbmuxd_response.parse_stream(sock)

        finally:
            # If we sent a bad request, we should re-create the socket in the correct version this time
            sock.close()
        sock = MuxConnection.create_usbmux_socket(usbmux_address=usbmux_address)

        if response.header.version == usbmuxd_version.BINARY:
            return BinaryMuxConnection(sock)
        elif response.header.version == usbmuxd_version.PLIST:
            return PlistMuxConnection(sock)

        raise MuxVersionError(f"usbmuxd returned unsupported version: {response.version}")

    def __init__(self, sock: SafeStreamSocket):
        self._sock = sock

        # after initiating the "Connect" packet, this same socket will be used to transfer data into the service
        # residing inside the target device. when this happens, we can no longer send/receive control commands to
        # usbmux on same socket
        self._connected = False

        # message sequence number. used when verifying the response matched the request
        self._tag = 1

        self.devices = []

    @abc.abstractmethod
    def _connect(self, device_id: int, port: int):
        """initiate a "Connect" request to target port"""
        pass

    @abc.abstractmethod
    def get_device_list(self, timeout: Optional[float] = None):
        """
        request an update to current device list
        """
        pass

    @abc.abstractmethod
    def listen(self):
        """start listening for events of attached and detached devices"""
        pass

    def connect(self, device: MuxDevice, port: int) -> socket.socket:
        """connect to a relay port on target machine and get a raw python socket object for the connection"""
        self._connect(device.devid, socket.htons(port))
        self._connected = True
        return self._sock.sock

    def close(self):
        """close current socket"""
        self._sock.close()

    def _assert_not_connected(self):
        """verify active state is in state for control messages"""
        if self._connected:
            raise MuxException("Mux is connected, cannot issue control packets")

    @abc.abstractmethod
    def receive_device_state_update(self):
        """Block and wait for a device notification"""
        pass

    def _raise_mux_exception(self, result: int, message: Optional[str] = None) -> None:
        exceptions = {
            int(usbmuxd_result.BADCOMMAND): BadCommandError,
            int(usbmuxd_result.BADDEV): BadDevError,
            int(usbmuxd_result.CONNREFUSED): ConnectionFailedError,
            int(usbmuxd_result.NOSUCHSERVICE): ConnectionFailedError,
            int(usbmuxd_result.BADVERSION): MuxVersionError,
        }
        exception = exceptions.get(result, MuxException)
        raise exception(message)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class BinaryMuxConnection(MuxConnection):
    """old binary protocol"""

    def __init__(self, sock: SafeStreamSocket):
        super().__init__(sock)
        self._version = usbmuxd_version.BINARY

    def get_device_list(self, timeout: Optional[float] = None):
        """use timeout to wait for the device list to be fully populated"""
        self._assert_not_connected()
        end = time.time() + timeout
        self.listen()
        while time.time() < end:
            self._sock.settimeout(end - time.time())
            try:
                self.receive_device_state_update()
            except (BlockingIOError, StreamError):
                continue
            except OSError as e:
                try:
                    self._sock.setblocking(True)
                    self.close()
                except OSError:
                    pass
                raise MuxException("Exception in listener socket") from e

    def listen(self):
        """start listening for events of attached and detached devices"""
        self._send_receive(usbmuxd_msgtype.LISTEN)

    def _connect(self, device_id: int, port: int):
        self._send({
            "header": {"version": self._version, "message": usbmuxd_msgtype.CONNECT, "tag": self._tag},
            "data": {"device_id": device_id, "port": port},
        })
        response = self._receive()
        if response.header.message != usbmuxd_msgtype.RESULT:
            raise MuxException(f"unexpected message type received: {response}")

        if response.data.result != usbmuxd_result.OK:
            raise self._raise_mux_exception(
                int(response.data.result),
                f"failed to connect to device: {device_id} at port: {port}. reason: {response.data.result}",
            )

    def _send(self, data: dict) -> None:
        self._assert_not_connected()
        self._sock.send(usbmuxd_request.build(data))
        self._tag += 1

    def _receive(self, expected_tag: Optional[int] = None):
        self._assert_not_connected()
        response = usbmuxd_response.parse_stream(self._sock)
        if expected_tag and response.header.tag != expected_tag:
            raise MuxException(f"Reply tag mismatch: expected {expected_tag}, got {response.header.tag}")
        return response

    def receive_device_state_update(self):
        response = self._receive()
        if response.header.message == usbmuxd_msgtype.ADD:
            # old protocol only supported USB devices
            self._add_device(MuxDevice(response.data.device_id, response.data.serial_number, "USB"))
        elif response.header.message == usbmuxd_msgtype.REMOVE:
            self._remove_device(response.data.device_id)
        else:
            raise MuxException(f"Invalid packet type received: {response}")

    def _send_receive(self, message_type: int):
        self._send({"header": {"version": self._version, "message": message_type, "tag": self._tag}, "data": b""})
        response = self._receive(self._tag - 1)
        if response.header.message != usbmuxd_msgtype.RESULT:
            raise MuxException(f"unexpected message type received: {response}")

        result = response.data.result
        if result != usbmuxd_result.OK:
            raise self._raise_mux_exception(int(result), f"{message_type} failed: error {result}")

    def _add_device(self, device: MuxDevice):
        self.devices.append(device)

    def _remove_device(self, device_id: int):
        self.devices = [device for device in self.devices if device.devid != device_id]


class PlistMuxConnection(BinaryMuxConnection):
    def __init__(self, sock: SafeStreamSocket):
        super().__init__(sock)
        self._version = usbmuxd_version.PLIST

    def listen(self) -> None:
        self._send_receive({"MessageType": "Listen"})

    def get_pair_record(self, serial: str) -> dict:
        # serials are saved inside usbmuxd without '-'
        self._send({"MessageType": "ReadPairRecord", "PairRecordID": serial})
        response = self._receive(self._tag - 1)
        pair_record = response.get("PairRecordData")
        if pair_record is None:
            raise NotPairedError("device should be paired first")
        return plistlib.loads(pair_record)

    def _process_device_state(self, response):
        if response["MessageType"] == "Attached":
            super()._add_device(
                MuxDevice(
                    response["DeviceID"],
                    response["Properties"]["SerialNumber"],
                    response["Properties"]["ConnectionType"],
                )
            )
        elif response["MessageType"] == "Detached":
            super()._remove_device(response["DeviceID"])
        else:
            raise MuxException(f"Invalid packet type received: {response}")

    def get_device_list(self, timeout: Optional[float] = None) -> None:
        """get device list synchronously without waiting the timeout"""
        self.devices = []
        self._send({"MessageType": "ListDevices"})
        response = self._receive(self._tag - 1)
        device_list = response.get("DeviceList")
        if device_list is None:
            raise MuxException(f"Got an invalid response from usbmux: {response}")
        for response in device_list:
            self._process_device_state(response)

    def get_buid(self) -> str:
        """get SystemBUID"""
        self._send({"MessageType": "ReadBUID"})
        return self._receive(self._tag - 1)["BUID"]

    def save_pair_record(self, serial: str, device_id: int, record_data: bytes):
        # serials are saved inside usbmuxd without '-'
        self._send_receive({
            "MessageType": "SavePairRecord",
            "PairRecordID": serial,
            "PairRecordData": record_data,
            "DeviceID": device_id,
        })

    def _connect(self, device_id: int, port: int):
        self._send_receive({"MessageType": "Connect", "DeviceID": device_id, "PortNumber": port})

    def _send(self, data: dict):
        request = {"ClientVersionString": "qt4i-usbmuxd", "ProgName": "pymobiledevice3", "kLibUSBMuxVersion": 3}
        request.update(data)
        super()._send({
            "header": {"version": self._version, "message": usbmuxd_msgtype.PLIST, "tag": self._tag},
            "data": plistlib.dumps(request),
        })

    def _receive(self, expected_tag: Optional[int] = None) -> dict:
        response = super()._receive(expected_tag=expected_tag)
        if response.header.message != usbmuxd_msgtype.PLIST:
            raise MuxException(f"Received non-plist type {response}")
        return plistlib.loads(response.data)

    def receive_device_state_update(self):
        response = self._receive()
        self._process_device_state(response)

    def _send_receive(self, data: dict):
        self._send(data)
        response = self._receive(self._tag - 1)
        if response["MessageType"] != "Result":
            raise MuxException(f"got an invalid message: {response}")
        if response["Number"] != 0:
            raise self._raise_mux_exception(response["Number"], f"got an error message: {response}")


def create_mux(usbmux_address: Optional[str] = None) -> MuxConnection:
    return MuxConnection.create(usbmux_address=usbmux_address)


def list_devices(usbmux_address: Optional[str] = None) -> list[MuxDevice]:
    mux = create_mux(usbmux_address=usbmux_address)
    try:
        mux.get_device_list(0.1)
        devices = mux.devices
    finally:
        mux.close()
    return devices


def select_device(
    udid: Optional[str] = None, connection_type: Optional[str] = None, usbmux_address: Optional[str] = None
) -> Optional[MuxDevice]:
    """
    select a UsbMux device according to given arguments.
    if more than one device could be selected, always prefer the usb one.
    """
    tmp = None
    for device in list_devices(usbmux_address=usbmux_address):
        if connection_type is not None and device.connection_type != connection_type:
            # if a specific connection_type was desired and not of this one then skip
            continue

        if udid is not None and not device.matches_udid(udid):
            # if a specific udid was desired and not of this one then skip
            continue

        # save best result as a temporary
        tmp = device

        if device.is_usb:
            # always prefer usb connection
            return device

    return tmp


def select_devices_by_connection_type(connection_type: str, usbmux_address: Optional[str] = None) -> list[MuxDevice]:
    """
    select all UsbMux devices by connection type
    """
    tmp = []
    for device in list_devices(usbmux_address=usbmux_address):
        if device.connection_type == connection_type:
            tmp.append(device)

    return tmp



import logging
import plistlib
import select
import socket
import struct
import threading
from enum import Enum

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import ConnectionFailedError, NoDeviceConnectedError, PyMobileDevice3Exception
from pymobiledevice3.service_connection import ServiceConnection

CTRL_PORT = 0x43A  # 1082
CTRLCMD = b"BeginCtrl\0"
HELLOCTRLCMD = b"HelloCtrl\0"
HELLOCMD = b"HelloConn\0"

FDR_SYNC_MSG = 0x1
FDR_PROXY_MSG = 0x105
FDR_PLIST_MSG = 0xBBAA
CHUNK_SIZE = 1048576

conn_port = None

logger = logging.getLogger(__name__)


class fdr_type(Enum):
    FDR_CTRL = 1
    FDR_CONN = 2


class FDRClient:
    SERVICE_PORT = CTRL_PORT

    ctrlprotoversion = 2

    def __init__(self, type_: fdr_type, udid=None):
        device = usbmux.select_device(udid)
        if device is None:
            if udid:
                raise ConnectionFailedError()
            else:
                raise NoDeviceConnectedError()

        logger.debug("connecting to FDR")

        if type_ == fdr_type.FDR_CTRL:
            self.service = ServiceConnection.create_using_usbmux(
                device.serial, self.SERVICE_PORT, connection_type="USB"
            )
            self.ctrl_handshake()
        else:
            self.service = ServiceConnection.create_using_usbmux(device.serial, conn_port, connection_type="USB")
            self.sync_handshake()

        logger.debug("FDR connected")

    def recv_plist(self):
        return self.service.recv_plist(endianity="<")

    def send_recv_plist(self, plist):
        return self.service.send_recv_plist(plist, endianity="<", fmt=plistlib.FMT_BINARY)

    def ctrl_handshake(self):
        global conn_port

        logger.debug("About to do ctrl handshake")

        self.service.sendall(CTRLCMD)

        if self.ctrlprotoversion != 2:
            raise NotImplementedError("TODO")

        req = {
            "Command": CTRLCMD,
            "CtrlProtoVersion": self.ctrlprotoversion,
        }
        resp = self.send_recv_plist(req)
        conn_port = resp["ConnPort"]

        logger.debug(f"Ctrl handshake done (ConnPort = {conn_port})")

    def sync_handshake(self):
        self.service.sendall(HELLOCMD)

        if self.ctrlprotoversion != 2:
            raise NotImplementedError("TODO")

        reply = self.recv_plist()
        cmd = reply["Command"]
        identifier = reply["Identifier"]

        if cmd != "HelloConn":
            raise PyMobileDevice3Exception("Did not receive HelloConn reply...")

        if identifier:
            logger.debug(f"got device identifier: {identifier}")

    def handle_sync_cmd(self):
        self.service.recvall(2)

        # Open a new connection and wait for messages on it
        logger.debug("FDR connected in reply to sync message, starting command thread")
        start_fdr_thread(fdr_type.FDR_CONN)

    def handle_proxy_cmd(self):
        buf = self.service.recv(1048576)
        logger.debug(f"got proxy command with {len(buf)} bytes")

        # Just return success here unconditionally because we don't know
        # anything else, and we will eventually abort on failure anyway
        self.service.sendall(struct.pack("<H", 5))

        if len(buf) < 3:
            logger.debug(f"FDR {self} proxy command data too short, retrying")
            return self.poll_and_handle_message()

        # ack command data too
        self.service.sendall(buf)

        host = None
        port = None

        # Now try to handle actual messages
        # Connect: 0 3 hostlen <host> <port>
        if buf[0] == 0 and buf[1] == 3:
            port = struct.unpack(">H", buf[-2:])[0]
            hostlen = buf[2]
            host = buf[3 : 3 + hostlen]

            logger.debug(f"FDR {self} Proxy connect request to {host}:{port}")

        if host is None:
            # missing or zero length host name
            return

        sockfd = socket.socket()
        sockfd.connect((host, port))

        while True:
            readable, _writable, exceptional = select.select(
                [sockfd, self.service.socket], [], [sockfd, self.service.socket]
            )

            for current_sock in readable:
                if current_sock == self.service.socket:
                    buf = self.service.recv(CHUNK_SIZE)

                    logger.debug(f"FDR {self} got payload of {len(buf)} bytes, now try to proxy it")
                    logger.debug(f"Sending {len(buf)} bytes of data")

                    sockfd.sendall(buf)
                else:
                    buf = sockfd.recv(CHUNK_SIZE)
                    logger.debug(f"Received {len(buf)} bytes")
                    self.service.sendall(buf)

            if exceptional:
                if exceptional[0] == sockfd:
                    logger.debug("Remote closed the connection")
                else:
                    logger.debug("Local service closed the connection")
                break

        sockfd.close()
        self.service.close()

    def handle_plist_cmd(self):
        d = self.recv_plist()
        command = d["Command"]

        if command == "Ping":
            self.send_recv_plist({"Pong": True})
        else:
            logger.warning(f"FDR {self} received unknown plist command: {command}")

    def poll_and_handle_message(self):
        # TODO: is it okay?
        cmd = struct.unpack("<H", self.service.recvall(2))[0]

        handlers = {
            FDR_SYNC_MSG: self.handle_sync_cmd,
            FDR_PROXY_MSG: self.handle_proxy_cmd,
            FDR_PLIST_MSG: self.handle_plist_cmd,
        }

        if cmd in handlers:
            handlers[cmd]()
        else:
            logger.warning(f"ignoring FDR message: {cmd}")


def fdr_listener_thread(type_: fdr_type):
    client = None
    try:
        client = FDRClient(type_)

        logger.debug(f"FDR {client} waiting for message...")

        while True:
            client.poll_and_handle_message()
    except ConnectionAbortedError:
        pass

    logger.debug(f"FDR {client} terminating...")


def start_fdr_thread(type_: fdr_type):
    thread = threading.Thread(target=fdr_listener_thread, args=(type_,))
    thread.start()
    return thread



import asyncio
import inspect
import logging
import re
from abc import ABC, abstractmethod
from asyncio import CancelledError
from collections.abc import AsyncIterator, Iterable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from functools import update_wrapper
from string import Template
from typing import Annotated, Any, Optional

import inquirer3
import IPython
import nest_asyncio
import typer
import uvicorn
from inquirer3.themes import GreenPassion
from prompt_toolkit import HTML, PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion.base import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document
from prompt_toolkit.history import FileHistory
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.styles import style_from_pygments_cls
from pygments import formatters, highlight, lexers
from pygments.styles import get_style_by_name
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep
from pymobiledevice3.common import get_home_folder
from pymobiledevice3.exceptions import (
    InspectorEvaluateError,
    LaunchingApplicationError,
    RemoteAutomationNotEnabledError,
    WebInspectorNotEnabledError,
    WirError,
)
from pymobiledevice3.lockdown import create_using_usbmux
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.services.web_protocol.cdp_server import app
from pymobiledevice3.services.web_protocol.driver import By, Cookie, WebDriver
from pymobiledevice3.services.web_protocol.inspector_session import InspectorSession
from pymobiledevice3.services.webinspector import SAFARI, Application, ApplicationPage, WebinspectorService

SCRIPT = Template("""
function inspectedPage_evalResult_getCompletions(primitiveType) {
    let resultSet = {};
    let object = primitiveType;
    for (let o = object; o; o = o.__proto__) {
        try {
            let names = Object.getOwnPropertyNames(o);
            for (let i = 0; i < names.length; ++i)
                resultSet[names[i]] = true;
        } catch(e) {}
    }
    return resultSet;
}

try {
    inspectedPage_evalResult_getCompletions(${object})
} catch (e) {}
""")

JS_RESERVED_WORDS = frozenset({
    "abstract",
    "arguments",
    "await",
    "boolean",
    "break",
    "byte",
    "case",
    "catch",
    "char",
    "class",
    "const",
    "continue",
    "debugger",
    "default",
    "delete",
    "do",
    "double",
    "else",
    "enum",
    "eval",
    "export",
    "extends",
    "false",
    "final",
    "finally",
    "float",
    "for",
    "function",
    "goto",
    "if",
    "implements",
    "import",
    "in",
    "instanceof",
    "int",
    "interface",
    "let",
    "long",
    "native",
    "new",
    "null",
    "package",
    "private",
    "protected",
    "public",
    "return",
    "short",
    "static",
    "super",
    "switch",
    "synchronized",
    "this",
    "throw",
    "throws",
    "transient",
    "true",
    "try",
    "typeof",
    "var",
    "void",
    "volatile",
    "while",
    "with",
    "yield",
})

OSUTILS = get_os_utils()
logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="webinspector",
    help=(
        "Control Safari/WebViews (tabs, automation, JS shells, CDP). "
        "Requires Web Inspector and Remote Automation enabled on the device."
    ),
    no_args_is_help=True,
)


def catch_errors(func):
    errors = {
        LaunchingApplicationError: "Unable to launch application (try to unlock device)",
        WebInspectorNotEnabledError: "Web inspector is not enabled",
        RemoteAutomationNotEnabledError: "Remote automation is not enabled",
    }

    def handle_error(e):
        logger.error(next(msg for exc, msg in errors.items() if isinstance(e, exc)))

    if inspect.iscoroutinefunction(func):

        async def catch_function(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except tuple(errors) as e:
                handle_error(e)

    else:

        def catch_function(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except tuple(errors) as e:
                handle_error(e)

    return update_wrapper(catch_function, func)


async def reload_pages(inspector: WebinspectorService) -> None:
    await inspector.get_open_pages()
    # Best effort.
    await inspector.flush_input(2)


async def create_webinspector_and_launch_app(
    lockdown: LockdownServiceProvider, timeout: float, app: str
) -> tuple[WebinspectorService, Application]:
    inspector = WebinspectorService(lockdown=lockdown)
    await inspector.connect(timeout)
    application = await inspector.open_app(app)
    return inspector, application


async def opened_tabs_task(service_provider: LockdownServiceProvider, timeout: float) -> None:
    inspector = WebinspectorService(lockdown=service_provider)
    await inspector.connect(timeout)
    application_pages = await inspector.get_open_application_pages(timeout=timeout)
    for application_page in application_pages:
        print(application_page)
    await inspector.close()


@cli.command()
@catch_errors
def opened_tabs(
    service_provider: ServiceProviderDep,
    timeout: Annotated[
        float,
        typer.Option("--timeout", "-t", help="Seconds to wait for WebInspector to respond."),
    ] = 3.0,
) -> None:
    """
    Show all currently opened tabs.

    \b
    Opt-in:
       iOS >= 18: Settings -> Apps -> Safari -> Advanced -> Web Inspector

       iOS < 18: Settings -> Safari -> Advanced -> Web Inspector
    """
    asyncio.run(opened_tabs_task(service_provider, timeout), debug=True)


@catch_errors
async def launch_task(service_provider: LockdownServiceProvider, url, timeout) -> None:
    inspector, safari = await create_webinspector_and_launch_app(service_provider, timeout, SAFARI)
    session = await inspector.automation_session(safari)
    driver = WebDriver(session)
    print("Starting session")
    await driver.start_session()
    print("Getting URL")
    await driver.get(url)
    OSUTILS.wait_return()
    await session.stop_session()
    await inspector.close()


@cli.command()
@catch_errors
def launch(
    service_provider: ServiceProviderDep,
    url: str,
    timeout: Annotated[
        float,
        typer.Option("--timeout", "-t", help="Seconds to wait for WebInspector to respond."),
    ] = 3.0,
) -> None:
    """
    Launch a specific URL in Safari.

    \b
    Opt-in (iOS >= 18):
        Settings -> Apps -> Safari -> Advanced -> Web Inspector
        Settings -> Apps -> Safari -> Advanced -> Remote Automation

    Opt-in (iOS < 18):
        Settings -> Safari -> Advanced -> Web Inspector
        Settings -> Safari -> Advanced -> Remote Automation

    """
    asyncio.run(launch_task(service_provider, url, timeout), debug=True)


SHELL_USAGE = """
# This shell allows you to control the web with selenium like API.
# The first thing you should do is creating a session:
driver.start_session()

# Then, you can navigate by using the get method:
driver.get("https://google.com")

# You can search for a specific element using the `By` enum:
driver.find_element(By.TAG_NAME, 'input')

# You can also add cookies:
driver.add_cookie(
    Cookie(name='tz', value='UTC', domain='.github.com', path='/', expires=0, httpOnly=False, secure=True,
    session=True, sameSite='None')
)

# See selenium api for more features.
"""


@catch_errors
async def shell_task(service_provider: LockdownServiceProvider, timeout: float) -> None:
    inspector, safari = await create_webinspector_and_launch_app(service_provider, timeout, SAFARI)
    session = await inspector.automation_session(safari)
    driver = WebDriver(session)
    try:
        nest_asyncio.apply()
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.Terminal256Formatter(style="native")),
            user_ns={
                "driver": driver,
                "Cookie": Cookie,
                "By": By,
            },
        )
    finally:
        await session.stop_session()
        await inspector.close()


@cli.command()
@catch_errors
def shell(
    service_provider: ServiceProviderDep,
    timeout: Annotated[
        float,
        typer.Option("--timeout", "-t", help="Seconds to wait for WebInspector to respond."),
    ] = 3.0,
) -> None:
    """
    Create an IPython shell for interacting with a WebView.

    \b
    Opt-in (iOS >= 18):
        Settings -> Apps -> Safari -> Advanced -> Web Inspector
        Settings -> Apps -> Safari -> Advanced -> Remote Automation

    Opt-in (iOS < 18):
        Settings -> Safari -> Advanced -> Web Inspector
        Settings -> Safari -> Advanced -> Remote Automation
    """
    asyncio.run(shell_task(service_provider, timeout), debug=True)


@cli.command()
@catch_errors
def js_shell(
    service_provider: ServiceProviderDep,
    url: str = "",
    timeout: Annotated[
        float,
        typer.Option("--timeout", "-t", help="Seconds to wait for WebInspector to respond."),
    ] = 3.0,
    automation: Annotated[
        bool,
        typer.Option(help="Use remote automation (requires Remote Automation toggle)."),
    ] = False,
    open_safari: Annotated[
        bool,
        typer.Option(help="Use an existing WebView; skip auto-opening Safari."),
    ] = False,
) -> None:
    """
    Create a javascript shell. This interpreter runs on your local machine,
    but evaluates each expression on the remote

    \b
    Opt-in:
        iOS >= 18: Settings -> Apps -> Safari -> Advanced -> Web Inspector
        iOS < 18: Settings -> Safari -> Advanced -> Web Inspector
    \b
    for automation also enable:
        iOS >= 18: Settings -> Apps -> Safari -> Advanced -> Remote Automation
        iOS < 18: Settings -> Safari -> Advanced -> Remote Automation
    """

    js_shell_class = AutomationJsShell if automation else InspectorJsShell
    asyncio.run(run_js_shell(js_shell_class, service_provider, timeout, url, open_safari))


udid = ""


def create_app():
    inspector = WebinspectorService(lockdown=create_using_usbmux(udid))
    app.state.inspector = inspector
    return app


@cli.command()
def cdp(service_provider: ServiceProviderDep, host: str = "127.0.0.1", port: int = 9222) -> None:
    """
    Start a CDP server for debugging WebViews.

    \b
    In order to debug the WebView that way, open in Google Chrome:
        chrome://inspect/#devices
    """
    global udid
    udid = service_provider.udid
    uvicorn.run(
        f"{__name__}:{create_app.__name__}",
        host=host,
        port=port,
        factory=True,
        ws_ping_timeout=None,
        ws="wsproto",
        loop="asyncio",
    )


async def get_js_completions(jsshell: "JsShell", obj: str, prefix: str) -> AsyncIterator[Completion]:
    if obj in JS_RESERVED_WORDS:
        return

    try:
        for key in await jsshell.evaluate_expression(SCRIPT.substitute(object=obj), return_by_value=True):
            if not key.startswith(prefix):
                continue
            yield Completion(key.removeprefix(prefix), display=key)
    except (Exception, CancelledError):
        # ignore every possible exception
        pass


class JsShellCompleter(Completer):
    def __init__(self, jsshell: "JsShell") -> None:
        self.jsshell: JsShell = jsshell

    async def get_completions_async(
        self,
        document: Document,
        complete_event: CompleteEvent,
    ) -> AsyncIterator[Completion]:
        # Build the JS expression we want to inspect
        text = f"globalThis.{document.text_before_cursor}"

        # Extract identifiers / dotted paths
        matches = re.findall(r"[a-zA-Z_][a-zA-Z_0-9.]+", text)
        if not matches:
            # async *generator*: just end, don't return a list
            return

        text = matches[-1]
        if "." in text:
            js_obj, prefix = text.rsplit(".", 1)
        else:
            js_obj = text
            prefix = ""

        # This should return an iterable of Completion (or something we can wrap)
        async for completion in get_js_completions(self.jsshell, js_obj, prefix):
            yield completion

    # Optional: keep sync completions empty so PTK knows we prefer async
    def get_completions(
        self,
        document: Document,
        complete_event: CompleteEvent,
    ) -> Iterable[Completion]:
        return []


class JsShell(ABC):
    def __init__(self) -> None:
        super().__init__()
        self.prompt_session: PromptSession = PromptSession(
            lexer=PygmentsLexer(lexers.JavascriptLexer),
            auto_suggest=AutoSuggestFromHistory(),
            style=style_from_pygments_cls(get_style_by_name("stata-dark")),
            history=FileHistory(self.webinspector_history_path()),
            completer=JsShellCompleter(self),
        )

    @classmethod
    @abstractmethod
    def create(
        cls, lockdown: LockdownServiceProvider, timeout: float, open_safari: bool
    ) -> "AbstractAsyncContextManager[JsShell]": ...

    @abstractmethod
    async def evaluate_expression(self, exp, return_by_value: bool = False) -> Any: ...

    @abstractmethod
    async def navigate(self, url: str) -> None: ...

    async def js_iter(self) -> None:
        with patch_stdout(True):
            exp = await self.prompt_session.prompt_async(HTML('<style fg="cyan"><b>&gt;</b></style> '))

        if not exp.strip():
            return

        result = await self.evaluate_expression(exp)
        colorful_result = highlight(
            f"{result}", lexers.JavascriptLexer(), formatters.Terminal256Formatter(style="stata-dark")
        )
        print(colorful_result, end="")

    async def start(self, url: str = "") -> None:
        if url:
            await self.navigate(url)
        while True:
            try:
                await self.js_iter()
            except (WirError, InspectorEvaluateError) as e:
                logger.error(e)
            except KeyboardInterrupt:  # KeyboardInterrupt Control-C
                pass
            except EOFError:  # Control-D
                return

    @staticmethod
    def webinspector_history_path() -> str:
        return str(get_home_folder() / "webinspector_history")


class AutomationJsShell(JsShell):
    def __init__(self, driver: WebDriver) -> None:
        super().__init__()
        self.driver: WebDriver = driver

    @classmethod
    @asynccontextmanager
    async def create(
        cls, lockdown: LockdownServiceProvider, timeout: float, open_safari: bool
    ) -> "AsyncIterator[AutomationJsShell]":
        inspector, application = await create_webinspector_and_launch_app(lockdown, timeout, SAFARI)
        automation_session = await inspector.automation_session(application)
        driver = WebDriver(automation_session)
        await driver.start_session()
        try:
            yield cls(driver)
        finally:
            await automation_session.stop_session()
            await inspector.close()

    async def evaluate_expression(self, exp: str, return_by_value: bool = False) -> Any:
        return await self.driver.execute_script(f"return {exp}")

    async def navigate(self, url: str) -> None:
        await self.driver.get(url)


class InspectorJsShell(JsShell):
    def __init__(self, inspector_session: InspectorSession) -> None:
        super().__init__()
        self.inspector_session: InspectorSession = inspector_session

    @classmethod
    @asynccontextmanager
    async def create(
        cls, lockdown: LockdownServiceProvider, timeout: float, open_safari: bool
    ) -> "AsyncIterator[InspectorJsShell]":
        inspector = WebinspectorService(lockdown=lockdown)
        await inspector.connect(timeout)
        if open_safari:
            _ = await inspector.open_app(SAFARI)
        application_page = await cls.query_page(inspector, bundle_identifier=SAFARI if open_safari else None)
        if application_page is None:
            raise typer.Exit()

        inspector_session = await inspector.inspector_session(application_page.application, application_page.page)
        await inspector_session.console_enable()
        await inspector_session.runtime_enable()

        try:
            yield cls(inspector_session)
        finally:
            await inspector.close()

    async def evaluate_expression(self, exp: str, return_by_value: bool = False) -> Any:
        return await self.inspector_session.runtime_evaluate(exp, return_by_value=return_by_value)

    async def navigate(self, url: str):
        await self.inspector_session.navigate_to_url(url)

    @staticmethod
    async def query_page(
        inspector: WebinspectorService, bundle_identifier: Optional[str] = None
    ) -> Optional[ApplicationPage]:
        available_pages = await inspector.get_open_application_pages(timeout=1)
        if bundle_identifier is not None:
            available_pages = [
                application_page
                for application_page in available_pages
                if application_page.application.bundle == bundle_identifier
            ]
        if not available_pages:
            logger.error("Unable to find available pages (try to unlock device)")
            return None

        page_query = [inquirer3.List("page", message="choose page", choices=available_pages, carousel=True)]
        page = inquirer3.prompt(page_query, theme=GreenPassion(), raise_keyboard_interrupt=True)["page"]
        return page


async def run_js_shell(
    js_shell_class: type[JsShell], lockdown: LockdownServiceProvider, timeout: float, url: str, open_safari: bool
) -> None:
    async with js_shell_class.create(lockdown, timeout, open_safari) as js_shell_instance:
        await js_shell_instance.start(url)



from io import BytesIO
from zipfile import ZIP_DEFLATED, ZipFile

import pytest

from pymobiledevice3.exceptions import AppInstallError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import AfcService
from pymobiledevice3.services.installation_proxy import InstallationProxyService


def _make_minimal_ipa_bytes() -> bytes:
    buffer = BytesIO()
    with ZipFile(buffer, "w", ZIP_DEFLATED) as zip_file:
        zip_file.writestr("Payload/Test.app/Info.plist", b"")
    return buffer.getvalue()


def test_get_apps(lockdown: LockdownClient) -> None:
    with InstallationProxyService(lockdown=lockdown) as installation_proxy:
        apps = installation_proxy.get_apps()
        assert len(apps) > 1


def test_get_system_apps(lockdown: LockdownClient) -> None:
    with InstallationProxyService(lockdown=lockdown) as installation_proxy:
        app_types = {app["ApplicationType"] for app in installation_proxy.get_apps(application_type="System").values()}
        assert len(app_types) == 1
        assert app_types.pop() == "System"


@pytest.mark.asyncio
async def test_parallel_installations_cleanup(lockdown: LockdownClient, monkeypatch: pytest.MonkeyPatch) -> None:
    ipa_bytes = _make_minimal_ipa_bytes()
    captured_plists: list[dict] = []

    def wrap_send_plist(service, captured: list[dict]):
        original_send_plist = service.send_plist

        def wrapped_send_plist(payload: dict) -> None:
            if payload.get("Command") == "Install":
                captured.append(payload.copy())
            original_send_plist(payload)

        return wrapped_send_plist

    with InstallationProxyService(lockdown=lockdown) as first, InstallationProxyService(lockdown=lockdown) as second:
        monkeypatch.setattr(first.service, "send_plist", wrap_send_plist(first.service, captured_plists))
        monkeypatch.setattr(second.service, "send_plist", wrap_send_plist(second.service, captured_plists))

        # TODO: perform those in parallel once ServiceConnection is thread safe (they read concurrently the "StartService" response)
        # results = await asyncio.gather(
        #     asyncio.to_thread(first.install_from_bytes, ipa_bytes),
        #     asyncio.to_thread(second.install_from_bytes, ipa_bytes),
        #     return_exceptions=True,
        # )
        try:
            result1 = first.install_from_bytes(ipa_bytes)
        except Exception as e:
            result1 = e
        try:
            result2 = second.install_from_bytes(ipa_bytes)
        except Exception as e:
            result2 = e
        results = [result1, result2]

    assert all(isinstance(result, AppInstallError) for result in results)
    assert len(captured_plists) == 2, f"Expected 2 captured plists, got {len(captured_plists)}"

    package_paths = [payload["PackagePath"] for payload in captured_plists]
    assert len(set(package_paths)) == 2, f"Expected different package paths for each installation, got {package_paths}"

    with AfcService(lockdown=lockdown) as afc:
        for package_path in package_paths:
            assert not afc.exists(package_path), f"Expected package path {package_path} to be cleaned up"



import ast
import asyncio
import logging
import plistlib
from pathlib import Path
from typing import Annotated, Literal, Optional

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import NoAutoPairServiceProviderDep, ServiceProviderDep, print_json, sudo_required
from pymobiledevice3.cli.remote import tunnel_task
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.common import TunnelProtocol
from pymobiledevice3.remote.tunnel_service import CoreDeviceTunnelProxy
from pymobiledevice3.services.heartbeat import HeartbeatService

logger = logging.getLogger(__name__)

cli = InjectingTyper(
    name="lockdown",
    help="Pair/Unpair device or access other lockdown services",
    no_args_is_help=True,
)


@cli.command("recovery")
def lockdown_recovery(service_provider: ServiceProviderDep) -> None:
    """enter recovery"""
    print_json(service_provider.enter_recovery())


@cli.command("service")
def lockdown_service(service_provider: ServiceProviderDep, service_name: str) -> None:
    """send-receive raw service messages with a given service name"""
    service_provider.start_lockdown_service(service_name).shell()


@cli.command("developer-service")
def lockdown_developer_service(service_provider: ServiceProviderDep, service_name: str) -> None:
    """send-receive raw service messages with a given developer service name"""
    service_provider.start_lockdown_developer_service(service_name).shell()


@cli.command("info")
def lockdown_info(service_provider: ServiceProviderDep) -> None:
    """query all lockdown values"""
    print_json(service_provider.all_values)


@cli.command("get")
def lockdown_get(service_provider: ServiceProviderDep, domain: Optional[str] = None, key: Optional[str] = None) -> None:
    """query lockdown values by their domain and key names"""
    print_json(service_provider.get_value(domain=domain, key=key))


@cli.command("set")
def lockdown_set(
    service_provider: ServiceProviderDep,
    value: str,
    domain: Optional[str] = None,
    key: Optional[str] = None,
) -> None:
    """set a lockdown value using python's ast.literal_eval()"""
    print_json(service_provider.set_value(value=ast.literal_eval(value), domain=domain, key=key))


@cli.command("remove")
def lockdown_remove(service_provider: ServiceProviderDep, domain: str, key: str) -> None:
    """remove a domain/key pair"""
    print_json(service_provider.remove_value(domain=domain, key=key))


@cli.command("unpair")
def lockdown_unpair(service_provider: NoAutoPairServiceProviderDep, host_id: Optional[str] = None) -> None:
    """unpair from connected device"""
    service_provider.unpair(host_id=host_id)


@cli.command("pair")
def lockdown_pair(service_provider: NoAutoPairServiceProviderDep) -> None:
    """pair device"""
    service_provider.pair()


@cli.command("pair-supervised")
def lockdown_pair_supervised(
    service_provider: NoAutoPairServiceProviderDep,
    keybag: Annotated[
        Path,
        typer.Argument(file_okay=True, dir_okay=False, exists=True),
    ],
) -> None:
    """pair supervised device"""
    service_provider.pair_supervised(keybag)


@cli.command("save-pair-record")
def lockdown_save_pair_record(service_provider: NoAutoPairServiceProviderDep, output: Path) -> None:
    """save pair record to specified location"""
    if service_provider.pair_record is None:
        logger.error("no pairing record was found")
        return
    output.write_bytes(plistlib.dumps(service_provider.pair_record))


@cli.command("date")
def lockdown_date(service_provider: ServiceProviderDep) -> None:
    """get device date"""
    print(service_provider.date)


@cli.command("heartbeat")
def lockdown_heartbeat(service_provider: ServiceProviderDep) -> None:
    """start heartbeat service"""
    HeartbeatService(service_provider).start()


@cli.command("language")
def lockdown_language(
    service_provider: ServiceProviderDep, language: Annotated[Optional[str], typer.Argument()] = None
) -> None:
    """Get/Set current language settings"""
    if language is not None:
        service_provider.set_language(language)
    print_json(service_provider.language)


@cli.command("locale")
def lockdown_locale(
    service_provider: ServiceProviderDep, locale: Annotated[Optional[str], typer.Argument()] = None
) -> None:
    """Get/Set current language settings"""
    if locale is not None:
        service_provider.set_locale(locale)
    print_json(service_provider.locale)


@cli.command("device-name")
def lockdown_device_name(service_provider: ServiceProviderDep, new_name: Optional[str] = None) -> None:
    """get/set current device name"""
    if new_name:
        service_provider.set_value(new_name, key="DeviceName")
    else:
        print(f"{service_provider.get_value(key='DeviceName')}")


@cli.command("wifi-connections")
def lockdown_wifi_connections(
    service_provider: ServiceProviderDep, state: Optional[Literal["on", "off"]] = None
) -> None:
    """get/set wifi connections state"""
    if not state:
        # show current state
        print_json(service_provider.get_value(domain="com.apple.mobile.wireless_lockdown"))
    else:
        # enable/disable
        service_provider.enable_wifi_connections = state == "on"


async def async_cli_start_tunnel(service_provider: LockdownServiceProvider, script_mode: bool) -> None:
    await tunnel_task(
        await CoreDeviceTunnelProxy.create(service_provider),
        script_mode=script_mode,
        secrets=None,
        protocol=TunnelProtocol.TCP,
    )


@cli.command("start-tunnel")
@sudo_required
def cli_start_tunnel(
    service_provider: ServiceProviderDep,
    script_mode: Annotated[
        bool,
        typer.Option(help="Show only HOST and port number to allow easy parsing from external shell scripts"),
    ] = False,
) -> None:
    """start tunnel"""
    asyncio.run(async_cli_start_tunnel(service_provider, script_mode), debug=True)


@cli.command("assistive-touch")
def lockdown_assistive_touch(
    service_provider: ServiceProviderDep, state: Optional[Literal["on", "off"]] = None
) -> None:
    """get/set assistive touch icon state (visibility)"""
    if not state:
        key = "AssistiveTouchEnabledByiTunes"
        accessibility_values = service_provider.get_value("com.apple.Accessibility")
        print_json({key: bool(accessibility_values[key])})
    else:
        # enable/disable
        service_provider.assistive_touch = state == "on"



import contextlib
import plistlib
from enum import Enum
from pathlib import Path
from typing import Any, Optional
from uuid import uuid4

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization.pkcs7 import PKCS7SignatureBuilder

from pymobiledevice3.exceptions import CloudConfigurationAlreadyPresentError, ProfileError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService

ERROR_CLOUD_CONFIGURATION_ALREADY_PRESENT = 14002
GLOBAL_HTTP_PROXY_UUID = "86a52338-52f7-4c09-b005-52baf3dc4882"
GLOBAL_RESTRICTIONS_UUID = "e22a0a66-08a8-43f5-9bbc-5279af35bb2b"


class Purpose(Enum):
    PostSetupInstallation = "PostSetupInstallation"


class MobileConfigService(LockdownService):
    SERVICE_NAME = "com.apple.mobile.MCInstall"
    RSD_SERVICE_NAME = "com.apple.mobile.MCInstall.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    def hello(self) -> None:
        self._send_recv({"RequestType": "HelloHostIdentifier"})

    def flush(self) -> None:
        self._send_recv({"RequestType": "Flush"})

    def escalate(self, keybag_file: Path) -> None:
        """
        Authenticate with the device.

        :param keybag_file: Certificate file in PEM format, containing certificate and private key.
        :return: None
        """
        with open(keybag_file, "rb") as keybag_file:
            keybag_file = keybag_file.read()
        private_key = serialization.load_pem_private_key(keybag_file, password=None)
        cer = x509.load_pem_x509_certificate(keybag_file)
        public_key = cer.public_bytes(Encoding.DER)
        escalate_response = self._send_recv({"RequestType": "Escalate", "SupervisorCertificate": public_key})
        signed_challenge = (
            PKCS7SignatureBuilder()
            .set_data(escalate_response["Challenge"])
            .add_signer(cer, private_key, hashes.SHA256())
            .sign(Encoding.DER, [])
        )
        self._send_recv({"RequestType": "EscalateResponse", "SignedRequest": signed_challenge})
        self._send_recv({"RequestType": "ProceedWithKeybagMigration"})

    def get_stored_profile(self, purpose: Purpose = Purpose.PostSetupInstallation) -> dict:
        return self._send_recv({"RequestType": "GetStoredProfile", "Purpose": purpose.value})

    def store_profile(self, profile_data: bytes, purpose: Purpose = Purpose.PostSetupInstallation) -> None:
        self._send_recv({"RequestType": "StoreProfile", "ProfileData": profile_data, "Purpose": purpose.value})

    def get_cloud_configuration(self) -> dict:
        return self._send_recv({"RequestType": "GetCloudConfiguration"}).get("CloudConfiguration")

    def set_cloud_configuration(self, cloud_configuration: dict) -> None:
        self._send_recv({"RequestType": "SetCloudConfiguration", "CloudConfiguration": cloud_configuration})

    def establish_provisional_enrollment(self, nonce: bytes) -> None:
        self._send_recv({"RequestType": "EstablishProvisionalEnrollment", "Nonce": nonce})

    def set_wifi_power_state(self, state: bool) -> None:
        self._send_recv({"RequestType": "SetWiFiPowerState", "PowerState": state})

    def erase_device(self, preserve_data_plan: bool, disallow_proximity_setup: bool) -> None:
        with contextlib.suppress(ConnectionAbortedError):
            self._send_recv({
                "RequestType": "EraseDevice",
                "PreserveDataPlan": preserve_data_plan,
                "DisallowProximitySetup": disallow_proximity_setup,
            })

    def get_profile_list(self) -> dict:
        return self._send_recv({"RequestType": "GetProfileList"})

    def install_profile(self, payload: bytes) -> None:
        self._send_recv({"RequestType": "InstallProfile", "Payload": payload})

    def install_profile_silent(self, keybag_file: Path, profile: bytes) -> None:
        self.escalate(keybag_file)
        self._send_recv({"RequestType": "InstallProfileSilent", "Payload": profile})

    def remove_profile(self, identifier: str) -> None:
        profiles = self.get_profile_list()
        if not profiles:
            return
        if identifier not in profiles["ProfileMetadata"]:
            self.logger.info(f"Trying to remove not installed profile: {identifier}")
            return
        meta = profiles["ProfileMetadata"][identifier]
        data = plistlib.dumps({
            "PayloadType": "Configuration",
            "PayloadIdentifier": identifier,
            "PayloadUUID": meta["PayloadUUID"],
            "PayloadVersion": meta["PayloadVersion"],
        })
        self._send_recv({"RequestType": "RemoveProfile", "ProfileIdentifier": data})

    def _send_recv(self, request: dict) -> dict:
        response = self.service.send_recv_plist(request)
        if response.get("Status", None) != "Acknowledged":
            error_chain = response.get("ErrorChain")
            if error_chain is not None:
                error_code = error_chain[0]["ErrorCode"]
                if error_code == ERROR_CLOUD_CONFIGURATION_ALREADY_PRESENT:
                    raise CloudConfigurationAlreadyPresentError()
            raise ProfileError(f"invalid response {response}")
        return response

    def install_wifi_profile(
        self,
        encryption_type: str,
        ssid: str,
        password: str,
        auto_join: bool = True,
        captive_bypass: bool = False,
        disable_association_mac_randomization: bool = False,
        hidden_network: bool = False,
        is_hotspot: bool = False,
        keybag_file: Optional[Path] = None,
    ) -> None:
        payload_uuid = str(uuid4())
        self.install_managed_profile(
            f"WiFi Profile For {ssid}",
            {
                "AutoJoin": auto_join,
                "CaptiveBypass": captive_bypass,
                "DisableAssociationMACRandomization": disable_association_mac_randomization,
                "EncryptionType": encryption_type,
                "HIDDEN_NETWORK": hidden_network,
                "IsHotspot": is_hotspot,
                "Password": password,
                "PayloadDescription": "Configures Wi-Fi settings",
                "PayloadDisplayName": "Wi-Fi",
                "PayloadIdentifier": f"com.apple.wifi.managed.{payload_uuid}",
                "PayloadType": "com.apple.wifi.managed",
                "PayloadUUID": payload_uuid,
                "PayloadVersion": 1,
                "ProxyType": "None",
                "SSID_STR": ssid,
            },
            keybag_file=keybag_file,
        )

    def install_http_proxy(self, server: str, server_port: int, keybag_file: Optional[Path] = None) -> None:
        payload_uuid = str(uuid4())
        self.install_managed_profile(
            f"HTTP Proxy for {server}:{server_port}",
            {
                "PayloadDescription": "Global HTTP Proxy",
                "PayloadDisplayName": "Global HTTP Proxy",
                "PayloadIdentifier": f"com.apple.proxy.http.global.{payload_uuid}",
                "PayloadType": "com.apple.proxy.http.global",
                "PayloadUUID": payload_uuid,
                "PayloadVersion": 1,
                "ProxyCaptiveLoginAllowed": False,
                "ProxyServer": server,
                "ProxyServerPort": server_port,
                "ProxyType": "Manual",
            },
            payload_uuid=GLOBAL_HTTP_PROXY_UUID,
            keybag_file=keybag_file,
        )

    def remove_http_proxy(self) -> None:
        self.remove_profile(GLOBAL_HTTP_PROXY_UUID)

    def supervise(self, organization: str, keybag_file: Path) -> None:
        cer = x509.load_pem_x509_certificate(keybag_file.read_bytes())
        public_key = cer.public_bytes(Encoding.DER)
        self.set_cloud_configuration({
            "AllowPairing": True,
            "CloudConfigurationUIComplete": True,
            "ConfigurationSource": 2,
            "ConfigurationWasApplied": True,
            "IsMDMUnremovable": False,
            "IsMandatory": True,
            "IsMultiUser": False,
            "IsSupervised": True,
            "OrganizationMagic": str(uuid4()),
            "OrganizationName": organization,
            "PostSetupProfileWasInstalled": True,
            "SkipSetup": [
                "Location",
                "Restore",
                "SIMSetup",
                "Android",
                "AppleID",
                "IntendedUser",
                "TOS",
                "Siri",
                "ScreenTime",
                "Diagnostics",
                "SoftwareUpdate",
                "Passcode",
                "Biometric",
                "Payment",
                "Zoom",
                "DisplayTone",
                "MessagingActivationUsingPhoneNumber",
                "HomeButtonSensitivity",
                "CloudStorage",
                "ScreenSaver",
                "TapToSetup",
                "Keyboard",
                "PreferredLanguage",
                "SpokenLanguage",
                "WatchMigration",
                "OnBoarding",
                "TVProviderSignIn",
                "TVHomeScreenSync",
                "Privacy",
                "TVRoom",
                "iMessageAndFaceTime",
                "AppStore",
                "Safety",
                "Multitasking",
                "ActionButton",
                "TermsOfAddress",
                "AccessibilityAppearance",
                "Welcome",
                "Appearance",
                "RestoreCompleted",
                "UpdateCompleted",
                "WiFi",
                "Display",
                "Tone",
                "LanguageAndLocale",
                "TouchID",
                "TrueToneDisplay",
                "FileVault",
                "iCloudStorage",
                "iCloudDiagnostics",
                "Registration",
                "DeviceToDeviceMigration",
                "UnlockWithWatch",
                "Accessibility",
                "All",
                "ExpressLanguage",
                "Language",
                "N/A",
                "Region",
                "Avatar",
                "DeviceProtection",
                "Key",
                "LockdownMode",
                "Wallpaper",
                "PrivacySubtitle",
                "SecuritySubtitle",
                "DataSubtitle",
                "AppleIDSubtitle",
                "AppearanceSubtitle",
                "PreferredLang",
                "OnboardingSubtitle",
                "AppleTVSubtitle",
                "Intelligence",
                "WebContentFiltering",
                "CameraButton",
                "AdditionalPrivacySettings",
                "EnableLockdownMode",
                "OSShowcase",
                "SafetyAndHandling",
                "Tips",
                "AgeBasedSafetySettings",
            ],
            "SupervisorHostCertificates": [public_key],
        })

    def install_managed_profile(
        self,
        display_name: str,
        payload_content: dict[str, Any],
        payload_uuid: str = str(uuid4()),
        keybag_file: Optional[Path] = None,
    ) -> None:
        profile_data = plistlib.dumps({
            "PayloadContent": [payload_content],
            "PayloadDisplayName": display_name,
            "PayloadIdentifier": payload_uuid,
            "PayloadRemovalDisallowed": False,
            "PayloadType": "Configuration",
            "PayloadUUID": payload_uuid,
            "PayloadVersion": 1,
        })
        if keybag_file is not None:
            self.install_profile_silent(keybag_file, profile_data)
        else:
            self.install_profile(profile_data)

    def install_restrictions_profile(
        self,
        enforced_software_update_delay: int = 0,
        payload_uuid: str = GLOBAL_RESTRICTIONS_UUID,
        keybag_file: Optional[Path] = None,
    ) -> None:
        self.install_managed_profile(
            "Restrictions",
            {
                "PayloadDescription": "Configures restrictions",
                "PayloadDisplayName": "Restrictions",
                "PayloadIdentifier": f"com.apple.applicationaccess.{payload_uuid}",
                "PayloadType": "com.apple.applicationaccess",
                "PayloadUUID": payload_uuid,
                "PayloadVersion": 1,
                "allowActivityContinuation": True,
                "allowAddingGameCenterFriends": True,
                "allowAirPlayIncomingRequests": True,
                "allowAirPrint": True,
                "allowAirPrintCredentialsStorage": True,
                "allowAirPrintiBeaconDiscovery": True,
                "allowAppCellularDataModification": True,
                "allowAppClips": True,
                "allowAppInstallation": True,
                "allowAppRemoval": True,
                "allowApplePersonalizedAdvertising": True,
                "allowAssistant": True,
                "allowAssistantWhileLocked": True,
                "allowAutoCorrection": True,
                "allowAutoUnlock": True,
                "allowAutomaticAppDownloads": True,
                "allowBluetoothModification": True,
                "allowBookstore": True,
                "allowBookstoreErotica": True,
                "allowCamera": True,
                "allowCellularPlanModification": True,
                "allowChat": True,
                "allowCloudBackup": True,
                "allowCloudDocumentSync": True,
                "allowCloudPhotoLibrary": True,
                "allowContinuousPathKeyboard": True,
                "allowDefinitionLookup": True,
                "allowDeviceNameModification": True,
                "allowDeviceSleep": True,
                "allowDictation": True,
                "allowESIMModification": True,
                "allowEnablingRestrictions": True,
                "allowEnterpriseAppTrust": True,
                "allowEnterpriseBookBackup": True,
                "allowEnterpriseBookMetadataSync": True,
                "allowEraseContentAndSettings": True,
                "allowExplicitContent": True,
                "allowFilesNetworkDriveAccess": True,
                "allowFilesUSBDriveAccess": True,
                "allowFindMyDevice": True,
                "allowFindMyFriends": True,
                "allowFingerprintForUnlock": True,
                "allowFingerprintModification": True,
                "allowGameCenter": True,
                "allowGlobalBackgroundFetchWhenRoaming": True,
                "allowInAppPurchases": True,
                "allowKeyboardShortcuts": True,
                "allowManagedAppsCloudSync": True,
                "allowMultiplayerGaming": True,
                "allowMusicService": True,
                "allowNews": True,
                "allowNotificationsModification": True,
                "allowOpenFromManagedToUnmanaged": True,
                "allowOpenFromUnmanagedToManaged": True,
                "allowPairedWatch": True,
                "allowPassbookWhileLocked": True,
                "allowPasscodeModification": True,
                "allowPasswordAutoFill": True,
                "allowPasswordProximityRequests": True,
                "allowPasswordSharing": True,
                "allowPersonalHotspotModification": True,
                "allowPhotoStream": True,
                "allowPredictiveKeyboard": True,
                "allowProximitySetupToNewDevice": True,
                "allowRadioService": True,
                "allowRemoteAppPairing": True,
                "allowRemoteScreenObservation": True,
                "allowSafari": True,
                "allowScreenShot": True,
                "allowSharedStream": True,
                "allowSpellCheck": True,
                "allowSpotlightInternetResults": True,
                "allowSystemAppRemoval": True,
                "allowUIAppInstallation": True,
                "allowUIConfigurationProfileInstallation": True,
                "allowUSBRestrictedMode": True,
                "allowUnpairedExternalBootToRecovery": False,
                "allowUntrustedTLSPrompt": True,
                "allowVPNCreation": True,
                "allowVideoConferencing": True,
                "allowVoiceDialing": True,
                "allowWallpaperModification": True,
                "allowiTunes": True,
                "enforcedSoftwareUpdateDelay": enforced_software_update_delay,
                "forceAirDropUnmanaged": False,
                "forceAirPrintTrustedTLSRequirement": False,
                "forceAssistantProfanityFilter": False,
                "forceAuthenticationBeforeAutoFill": False,
                "forceAutomaticDateAndTime": False,
                "forceClassroomAutomaticallyJoinClasses": False,
                "forceClassroomRequestPermissionToLeaveClasses": False,
                "forceClassroomUnpromptedAppAndDeviceLock": False,
                "forceClassroomUnpromptedScreenObservation": False,
                "forceDelayedSoftwareUpdates": True,
                "forceEncryptedBackup": False,
                "forceITunesStorePasswordEntry": False,
                "forceLimitAdTracking": False,
                "forceWatchWristDetection": False,
                "forceWiFiPowerOn": False,
                "forceWiFiWhitelisting": False,
                "ratingApps": 1000,
                "ratingMovies": 1000,
                "ratingRegion": "us",
                "ratingTVShows": 1000,
                "safariAcceptCookies": 2.0,
                "safariAllowAutoFill": True,
                "safariAllowJavaScript": True,
                "safariAllowPopups": True,
                "safariForceFraudWarning": False,
            },
            payload_uuid=payload_uuid,
            keybag_file=keybag_file,
        )



import logging
import threading
import time
from typing import Any, Optional

from bpylist2 import archiver
from packaging.version import Version

from pymobiledevice3.exceptions import AppNotInstalledError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.dvt_testmanaged_proxy import DvtTestmanagedProxyService
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl
from pymobiledevice3.services.house_arrest import HouseArrestService
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.services.remote_server import (
    NSURL,
    NSUUID,
    Channel,
    MessageAux,
    RemoteServer,
    XCTCapabilities,
    XCTestConfiguration,
)

logger = logging.getLogger(__name__)


class ReverseRemoteService:
    def __init__(self, service: RemoteServer, channel: int):
        self.service = service
        self.channel = channel
        self.logger = logger.getChild(self.__class__.__name__)

    def __on_method_call__(self, selector: str, args: Optional[MessageAux] = None):
        _selector = selector.replace(":", "_")
        if not hasattr(self, _selector):
            logger.debug(
                "%s: unhandled method call: selector='%s', _selector='%s', args='%s'",
                self.__class__.__name__,
                selector,
                _selector,
                args,
            )
            self.service.send_reply_error(self.channel, self.service.cur_remote_message)
            return
        ret = getattr(self, _selector)(args)
        self.service.send_reply(self.channel, self.service.cur_remote_message, ret)


class XCUITestListener(ReverseRemoteService):
    """dtxproxy:XCTestDriverInterface:XCTestManager_IDEInterface"""

    def __init__(self, service: DvtTestmanagedProxyService, channel: int, test_configuration: XCTestConfiguration):
        super().__init__(service, channel)
        self.test_config = test_configuration

    def __on_method_call__(self, selector: str, args: Optional[MessageAux] = None):
        return super().__on_method_call__(selector.removeprefix("_XCT_"), args)

    def logDebugMessage_(self, args: Optional[MessageAux] = None):
        self.logger.debug("logDebugMessage: %s", args[0].value.strip())

    def testRunnerReadyWithCapabilities_(self, args: Optional[MessageAux] = None):
        self.logger.info("testRunnerReadyWithCapabilities: %s", args[0].value)
        return self.test_config

    def didFinishExecutingTestPlan(self, args: Optional[MessageAux] = None):
        self.logger.info("didFinishExecutingTestPlan")


class XCUITestPlanConsumer:
    def __init__(
        self,
        pid: int,
        pctrl: ProcessControl,
        ctrl_dvt: DvtTestmanagedProxyService,
        ctrl_chan: Channel,
        main_dvt: DvtTestmanagedProxyService,
        main_chan: Channel,
        test_configuration: XCTestConfiguration,
    ):
        self.pid = pid
        self.pctrl = pctrl
        self.ctrl_dvt = ctrl_dvt
        self.ctrl_chan = ctrl_chan
        self.main_dvt = main_dvt
        self.main_chan = main_chan
        self.xctest_config = test_configuration
        self.listener: Optional[XCUITestListener] = None
        self.__running__ = False
        self.__closing__ = False
        self.completed = threading.Event()

    def set_listener(self, listener: XCUITestListener):
        self.listener = listener

    def consume(self):
        if self.listener is None:
            self.listener = XCUITestListener(self.main_dvt, self.main_chan, self.xctest_config)
        try:
            self.__running__ = True
            while self.__running__:
                key, value = self.main_dvt.recv_plist(self.main_chan)
                if key == "_XCT_didFinishExecutingTestPlan":
                    self.__closing__ = True
                self.listener.__on_method_call__(key, value)
        except (ConnectionAbortedError, ConnectionResetError):
            if not self.__closing__:
                logger.warning("connection aborted")
            return
        except (InterruptedError, KeyboardInterrupt):
            return
        finally:
            self.stop()
            self.completed.set()

    def stop(self):
        if not self.__running__:
            return

        self.__closing__ = True
        self.__running__ = False

        logger.info("Killing UITest with pid %d ...", self.pid)
        self.pctrl.kill(self.pid)
        self.main_dvt.close()
        self.ctrl_dvt.close()


class XCUITestService:
    IDENTIFIER = "dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface"
    XCODE_VERSION = 36  # not important
    __TESTDRIVERINTERFACE_CHANNEL__ = -1  # a special channel that is reserved for "XCTestDriverInterface"

    def __init__(self, service_provider: LockdownServiceProvider):
        self.service_provider = service_provider
        self.pctl = self.init_process_control()
        self.product_major_version = Version(service_provider.product_version).major

    def start(
        self,
        bundle_id: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ) -> XCUITestPlanConsumer:
        session_identifier = NSUUID.uuid4()
        app_info = get_app_info(self.service_provider, bundle_id)

        xctest_configuration = generate_xctestconfiguration(
            app_info, session_identifier, self.product_major_version, test_runner_env, test_runner_args
        )
        xctest_path = f"/tmp/{str(session_identifier).upper()}.xctestconfiguration"  # yapf: disable

        if self.product_major_version < 17:
            self.setup_xcuitest(bundle_id, xctest_path, xctest_configuration)

        ctrl_dvt, ctrl_chan, main_dvt, main_chan = self.init_ide_channels(
            session_identifier, xctest_configuration._config["IDECapabilities"]
        )

        pid = self.launch_test_app(
            session_identifier, app_info, bundle_id, xctest_path, test_runner_env, test_runner_args
        )
        logger.info("Runner started with pid:%d, waiting for testBundleReady", pid)

        if self.product_major_version < 17:
            time.sleep(1)

        self.authorize_test_process_id(ctrl_chan, pid)
        # acknoledge that we will have this service listening on our channel
        main_dvt.serve_channel("dtxproxy:XCTestDriverInterface:XCTestManager_IDEInterface", main_chan)
        self.start_executing_test_plan_with_protocol_version(main_dvt, self.XCODE_VERSION)

        # TODO: boradcast message is not handled
        # TODO: RemoteServer.receive_message is not thread safe and will block if no message received
        return XCUITestPlanConsumer(pid, self.pctl, ctrl_dvt, ctrl_chan, main_dvt, main_chan, xctest_configuration)

    def run(
        self,
        bundle_id: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ):
        consumer = self.start(bundle_id, test_runner_env, test_runner_args)
        consumer.consume()

    def init_process_control(self):
        dvt_proxy = DvtSecureSocketProxyService(lockdown=self.service_provider)
        dvt_proxy.perform_handshake()
        return ProcessControl(dvt_proxy)

    def init_ide_channels(self, session_identifier: NSUUID, ide_capabilities: XCTCapabilities):
        """
        Initialize the two channels required by XcodeIDE to control the test session. The first channel is used to authorize the test process, and the second channel is used to manage the test.

        return: (ctrl_dvt, ctrl_chan, main_dvt, main_chan)
         - ctrl_dvt: the DvtTestmanagedProxyService instance for the control channel
         - ctrl_chan: the Channel instance for the control channel
         - main_dvt: the DvtTestmanagedProxyService instance for the main channel
         - main_chan: the Channel instance for the main channel
        """
        # XcodeIDE require two connections
        ctrl_dvt = DvtTestmanagedProxyService(lockdown=self.service_provider)
        ctrl_dvt.perform_handshake()

        logger.info("make channel %s", self.IDENTIFIER)
        ctrl_chan = ctrl_dvt.make_channel(self.IDENTIFIER)
        if self.product_major_version >= 17:
            ctrl_dvt.send_message(
                ctrl_chan,
                "_IDE_initiateControlSessionWithCapabilities:",
                MessageAux().append_obj(XCTCapabilities({})),
            )
            reply = ctrl_chan.receive_plist()
            logger.info("ctrl_conn handshake capabilities: %s", reply)
        elif self.product_major_version >= 11:
            ctrl_dvt.send_message(
                ctrl_chan,
                "_IDE_initiateControlSessionWithProtocolVersion:",
                MessageAux().append_obj(self.XCODE_VERSION),
            )
            reply = ctrl_chan.receive_plist()
            logger.info("ctrl_conn handshake xcode version: %s", reply)

        main_dvt = DvtTestmanagedProxyService(lockdown=self.service_provider)
        main_dvt.perform_handshake()
        main_chan = main_dvt.make_channel(self.IDENTIFIER)
        if self.product_major_version >= 17:
            main_dvt.send_message(
                channel=main_chan,
                selector="_IDE_initiateSessionWithIdentifier:capabilities:",
                args=MessageAux().append_obj(session_identifier).append_obj(ide_capabilities),
            )
            reply = main_chan.receive_plist()
            logger.info("main_conn handshake capabilities: %s", reply)
        else:
            main_dvt.send_message(
                channel=main_chan,
                selector="_IDE_initiateSessionWithIdentifier:forClient:atPath:protocolVersion:",
                args=MessageAux()
                .append_obj(session_identifier)
                .append_obj("not-very-import-part")  # this part is not important
                .append_obj("/Applications/Xcode.app/Contents/Developer/usr/bin/xcodebuild")
                .append_obj(self.XCODE_VERSION),
            )
            reply = main_chan.receive_plist()
            logger.info("main_conn handshake xcode version: %s", reply)
        return ctrl_dvt, ctrl_chan, main_dvt, main_chan

    def setup_xcuitest(
        self,
        bundle_id: str,
        xctest_path: str,
        xctest_configuration: XCTestConfiguration,
    ):
        """push xctestconfiguration to app VendDocuments"""
        with HouseArrestService(lockdown=self.service_provider, bundle_id=bundle_id, documents_only=False) as afc:
            for name in afc.listdir("/tmp"):
                if name.endswith(".xctestconfiguration"):
                    logger.debug("remove /tmp/%s", name)
                    afc.rm("/tmp/" + name)
            afc.set_file_contents(xctest_path, archiver.archive(xctest_configuration))

    def start_executing_test_plan_with_protocol_version(self, dvt: DvtTestmanagedProxyService, protocol_version: int):
        dvt.send_message(
            self.__TESTDRIVERINTERFACE_CHANNEL__,
            "_IDE_startExecutingTestPlanWithProtocolVersion:",
            MessageAux().append_obj(protocol_version),
            expects_reply=False,
        )

    def authorize_test_process_id(self, chan: Channel, pid: int):
        selector = None
        aux = MessageAux()
        if self.product_major_version >= 12:
            selector = "_IDE_authorizeTestSessionWithProcessID:"
            aux.append_obj(pid)
        elif self.product_major_version >= 10:
            selector = "_IDE_initiateControlSessionForTestProcessID:protocolVersion:"
            aux.append_obj(pid)
            aux.append_obj(self.XCODE_VERSION)
        else:
            selector = "_IDE_initiateControlSessionForTestProcessID:"
            aux.append_obj(pid)
        chan.send_message(selector, aux)
        reply = chan.receive_plist()
        if isinstance(reply, bool) and reply is True:
            logger.info("authorizing test session for pid %d successful %r", pid, reply)
        else:
            raise RuntimeError(f"Failed to authorize test process id: {reply}")

    def launch_test_app(
        self,
        test_session_identifier: NSUUID,
        app_info: dict,
        bundle_id: str,
        xctest_path: str,
        test_runner_env: Optional[dict] = None,
        test_runner_args: Optional[list] = None,
    ) -> int:
        app_container = app_info["Container"]
        app_path = app_info["Path"]
        exec_name = app_info["CFBundleExecutable"]
        # # logger.info('CFBundleExecutable: %s', exec_name)
        # # CFBundleName always endswith -Runner
        assert exec_name.endswith("-Runner"), f"Invalid CFBundleExecutable: {exec_name}"
        target_name = exec_name[: -len("-Runner")]

        app_env = {
            "CA_ASSERT_MAIN_THREAD_TRANSACTIONS": "0",
            "CA_DEBUG_TRANSACTIONS": "0",
            "DYLD_FRAMEWORK_PATH": app_path + "/Frameworks:",
            "DYLD_LIBRARY_PATH": app_path + "/Frameworks",
            "MTC_CRASH_ON_REPORT": "1",
            "NSUnbufferedIO": "YES",
            "SQLITE_ENABLE_THREAD_ASSERTIONS": "1",
            "WDA_PRODUCT_BUNDLE_IDENTIFIER": "",
            "XCTestBundlePath": f"{app_info['Path']}/PlugIns/{target_name}.xctest",
            "XCTestConfigurationFilePath": app_container + xctest_path,
            "XCODE_DBG_XPC_EXCLUSIONS": "com.apple.dt.xctestSymbolicator",
            # the following maybe no needed
            # 'MJPEG_SERVER_PORT': '',
            # 'USE_PORT': '',
            # 'LLVM_PROFILE_FILE': app_container + '/tmp/%p.profraw', # %p means pid
            "XCTestSessionIdentifier": str(test_session_identifier).upper(),
        }

        if self.product_major_version >= 11:
            app_env["DYLD_INSERT_LIBRARIES"] = "/Developer/usr/lib/libMainThreadChecker.dylib"
            app_env["OS_ACTIVITY_DT_MODE"] = "YES"
        if self.product_major_version >= 17:
            app_env["DYLD_FRAMEWORK_PATH"] = f"${app_env['DYLD_FRAMEWORK_PATH']}/System/Developer/Library/Frameworks:"
            app_env["DYLD_LIBRARY_PATH"] = f"${app_env['DYLD_LIBRARY_PATH']}:/System/Developer/usr/lib"
            app_env["XCTestConfigurationFilePath"] = ""  # sent as return value of _XCT_testRunnerReadyWithCapabilities
            app_env["XCTestManagerVariant"] = "DDI"

        if test_runner_env:
            app_env.update(test_runner_env)

        app_args = [
            "-NSTreatUnknownArgumentsAsOpen",
            "NO",
            "-ApplePersistenceIgnoreState",
            "YES",
        ]
        app_args.extend(test_runner_args or [])
        app_options = {"StartSuspendedKey": False}
        if self.product_major_version >= 12:
            app_options["ActivateSuspended"] = True

        # go-ios uses the following to spawn a new process for the test runner.
        # async def launch_app_async():
        #     async with AppServiceService(self.service_provider) as app_service:
        #         resp = await app_service.launch_application(bundle_id, arguments=app_args, environment=app_env, extra_options=app_options, kill_existing=True)
        #         return int(resp["processToken"]["processIdentifier"])
        # pid = asyncio.run(launch_app_async())
        return self.pctl.launch(
            bundle_id,
            arguments=app_args,
            environment=app_env,
            extra_options=app_options,
        )


def get_app_info(service_provider: LockdownClient, bundle_id: str) -> dict[str, Any]:
    with InstallationProxyService(lockdown=service_provider) as install_service:
        apps = install_service.get_apps(bundle_identifiers=[bundle_id])
        if not apps:
            raise AppNotInstalledError(f"No app with bundle id {bundle_id} found")
        return apps[bundle_id]


def generate_xctestconfiguration(
    target_app_info: dict,
    session_identifier: NSUUID,
    product_major_version: int,
    target_app_env: Optional[dict] = None,
    target_app_args: Optional[list] = None,
    tests_to_run: Optional[list] = None,
    tests_to_skip: Optional[list] = None,
) -> XCTestConfiguration:
    exec_name: str = target_app_info["CFBundleExecutable"]
    assert exec_name.endswith("-Runner"), f"Invalid CFBundleExecutable: {exec_name}"
    config_name = exec_name[: -len("-Runner")]

    automation_framework_path = "/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework"
    if product_major_version >= 17:
        automation_framework_path = "/System/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework"

    return XCTestConfiguration({
        "testBundleURL": NSURL(None, f"file://{target_app_info['Path']}/PlugIns/{config_name}.xctest"),
        "sessionIdentifier": session_identifier,
        "productModuleName": config_name,
        "targetApplicationBundleID": target_app_info.get("CFBundleIdentifier"),
        "targetApplicationPath": target_app_info.get("Path"),
        "targetApplicationEnvironment": target_app_env or {},
        "targetApplicationArguments": target_app_args or [],
        "testsToRun": tests_to_run or set(),
        "testsToSkip": tests_to_skip or set(),
        "testsMustRunOnMainThread": True,
        "reportResultsToIDE": True,
        "reportActivities": True,
        "automationFrameworkPath": automation_framework_path,
        "testApplicationDependencies": None,
    })



from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService


class Screenshot:
    IDENTIFIER = "com.apple.instruments.server.services.screenshot"

    def __init__(self, dvt: DvtSecureSocketProxyService):
        self._channel = dvt.make_channel(self.IDENTIFIER)

    def get_screenshot(self) -> bytes:
        """get device screenshot"""
        self._channel.takeScreenshot(expects_reply=True)
        return self._channel.receive_plist()



import contextlib
import copy
import io
import logging
import os
import plistlib
import threading
import uuid
from contextlib import suppress
from functools import partial
from pprint import pprint
from queue import Empty, Queue
from typing import ClassVar, Optional

import IPython
from bpylist2 import archiver
from construct import (
    Adapter,
    Const,
    Container,
    Default,
    GreedyBytes,
    GreedyRange,
    Int16ul,
    Int32sl,
    Int32ul,
    Int64ul,
    Prefixed,
    Select,
    Struct,
    Switch,
    this,
)
from pygments import formatters, highlight, lexers

from pymobiledevice3.exceptions import DvtException, UnrecognizedSelectorError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService

SHELL_USAGE = """
# This shell allows you to send messages to the DVTSecureSocketProxy and receive answers easily.
# Generally speaking, each channel represents a group of actions.
# Calling actions is done using a selector and auxiliary (parameters).
# Receiving answers is done by getting a return value and seldom auxiliary (private / extra parameters).
# To see the available channels, type the following:
developer.supported_identifiers

# In order to send messages, you need to create a channel:
channel = developer.make_channel('com.apple.instruments.server.services.deviceinfo')

# After creating the channel you can call allowed selectors:
channel.runningProcesses()

# If an answer is expected, you can receive it using the receive method:
processes = channel.receive_plist()

# You can also call methods from the broadcast channel
broadcast.someMethod()

# Sometimes the selector requires parameters, You can add them using MessageAux. For example lets kill a process:
channel = developer.make_channel('com.apple.instruments.server.services.processcontrol')
args = MessageAux().append_obj(80) # This will kill pid 80
channel.killPid_(args, expects_reply=False) # Killing a process doesn't require an answer.

# In some rare cases, you might want to receive the auxiliary and the selector return value.
# For that cases you can use the recv_plist method.
return_value, auxiliary = developer.recv_plist()
"""


class BplitAdapter(Adapter):
    def _decode(self, obj, context, path):
        return archiver.unarchive(obj)

    def _encode(self, obj, context, path):
        return archiver.archive(obj)


message_aux_t_struct = Struct(
    "magic" / Default(Int64ul, 0x1F0),
    "aux"
    / Prefixed(
        Int64ul,
        GreedyRange(
            Struct(
                "_empty_dictionary" / Select(Const(0xA, Int32ul), Int32ul),
                "type" / Int32ul,
                "value"
                / Switch(
                    this.type,
                    {2: BplitAdapter(Prefixed(Int32ul, GreedyBytes)), 3: Int32ul, 6: Int64ul},
                    default=GreedyBytes,
                ),
            )
        ),
    ),
)
dtx_message_header_struct = Struct(
    "magic" / Const(0x1F3D5B79, Int32ul),
    "cb" / Int32ul,
    "fragmentId" / Int16ul,
    "fragmentCount" / Int16ul,
    "length" / Int32ul,
    "identifier" / Int32ul,
    "conversationIndex" / Int32ul,
    "channelCode" / Int32sl,
    "expectsReply" / Int32ul,
)
dtx_message_payload_header_struct = Struct(
    "flags" / Int32ul,
    "auxiliaryLength" / Int32ul,
    "totalLength" / Int64ul,
)


class MessageAux:
    def __init__(self):
        self.values = []

    def append_int(self, value: int):
        self.values.append({"type": 3, "value": value})
        return self

    def append_long(self, value: int):
        self.values.append({"type": 6, "value": value})
        return self

    def append_obj(self, value):
        self.values.append({"type": 2, "value": value})
        return self

    def __bytes__(self):
        return message_aux_t_struct.build({"aux": self.values})


class DTTapMessage:
    @staticmethod
    def decode_archive(archive_obj):
        return archive_obj.decode("DTTapMessagePlist")


class NSNull:
    @staticmethod
    def decode_archive(archive_obj):
        return None


class NSError:
    @staticmethod
    def encode_archive(archive_obj):
        return archiver.archive(archive_obj)

    @staticmethod
    def decode_archive(archive_obj):
        user_info = archive_obj.decode("NSUserInfo")
        if user_info.get("NSLocalizedDescription", "").endswith(" - it does not respond to the selector"):
            raise UnrecognizedSelectorError(user_info)
        raise DvtException(archive_obj.decode("NSUserInfo"))


class NSUUID(uuid.UUID):
    @staticmethod
    def uuid4():
        """Generate a random UUID."""
        return NSUUID(bytes=os.urandom(16))

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        archive_obj.encode("NS.uuidbytes", self.bytes)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return NSUUID(bytes=archive_obj.decode("NS.uuidbytes"))


class NSURL:
    def __init__(self, base, relative):
        self.base = base
        self.relative = relative

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        archive_obj.encode("NS.base", self.base)
        archive_obj.encode("NS.relative", self.relative)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return NSURL(archive_obj.decode("NS.base"), archive_obj.decode("NS.relative"))


class NSValue:
    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.decode("NS.rectval")


class NSMutableData:
    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.decode("NS.data")


class NSMutableString:
    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.decode("NS.string")


class XCTCapabilities:
    def __init__(self, capabilities: dict):
        self.capabilities = capabilities

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        archive_obj.encode("capabilities-dictionary", self.capabilities)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return XCTCapabilities(archive_obj.decode("capabilities-dictionary"))

    def __str__(self):
        return f"XCTCapabilities({self.capabilities})"


class XCTestConfiguration:
    _default: ClassVar = {
        # 'testBundleURL': UID(3),
        # 'sessionIdentifier': UID(8), # UUID
        "aggregateStatisticsBeforeCrash": {"XCSuiteRecordsKey": {}},
        "automationFrameworkPath": "/Developer/Library/PrivateFrameworks/XCTAutomationSupport.framework",
        "baselineFileRelativePath": None,
        "baselineFileURL": None,
        "defaultTestExecutionTimeAllowance": None,
        "disablePerformanceMetrics": False,
        "emitOSLogs": False,
        "formatVersion": plistlib.UID(2),  # store in UID
        "gatherLocalizableStringsData": False,
        "initializeForUITesting": True,
        "maximumTestExecutionTimeAllowance": None,
        "productModuleName": "WebDriverAgentRunner",  # set to other value is also OK
        "randomExecutionOrderingSeed": None,
        "reportActivities": True,
        "reportResultsToIDE": True,
        "systemAttachmentLifetime": 2,
        "targetApplicationArguments": [],  # maybe useless
        "targetApplicationBundleID": None,
        "targetApplicationEnvironment": None,
        "targetApplicationPath": "/whatever-it-does-not-matter/but-should-not-be-empty",
        "testApplicationDependencies": {},
        "testApplicationUserOverrides": None,
        "testBundleRelativePath": None,
        "testExecutionOrdering": 0,
        "testTimeoutsEnabled": False,
        "testsDrivenByIDE": False,
        "testsMustRunOnMainThread": True,
        "testsToRun": None,
        "testsToSkip": None,
        "treatMissingBaselinesAsFailures": False,
        "userAttachmentLifetime": 0,
        "preferredScreenCaptureFormat": 2,
        "IDECapabilities": XCTCapabilities({
            "expected failure test capability": True,
            "test case run configurations": True,
            "test timeout capability": True,
            "test iterations": True,
            "request diagnostics for specific devices": True,
            "delayed attachment transfer": True,
            "skipped test capability": True,
            "daemon container sandbox extension": True,
            "ubiquitous test identifiers": True,
            "XCTIssue capability": True,
        }),
    }

    def __init__(self, kv: dict):
        assert "testBundleURL" in kv
        assert "sessionIdentifier" in kv
        self._config = copy.deepcopy(self._default)
        self._config.update(kv)

    def encode_archive(self, archive_obj: archiver.ArchivingObject):
        for k, v in self._config.items():
            archive_obj.encode(k, v)

    @staticmethod
    def decode_archive(archive_obj: archiver.ArchivedObject):
        return archive_obj.object


archiver.update_class_map({
    "DTSysmonTapMessage": DTTapMessage,
    "DTTapHeartbeatMessage": DTTapMessage,
    "DTTapStatusMessage": DTTapMessage,
    "DTKTraceTapMessage": DTTapMessage,
    "DTActivityTraceTapMessage": DTTapMessage,
    "DTTapMessage": DTTapMessage,
    "NSNull": NSNull,
    "NSError": NSError,
    "NSUUID": NSUUID,
    "NSURL": NSURL,
    "NSValue": NSValue,
    "NSMutableData": NSMutableData,
    "NSMutableString": NSMutableString,
    "XCTestConfiguration": XCTestConfiguration,
    "XCTCapabilities": XCTCapabilities,
})

archiver.Archive.inline_types = list({*archiver.Archive.inline_types, bytes})


class Channel(int):
    @classmethod
    def create(cls, value: int, service: "RemoteServer"):
        channel = cls(value)
        channel._service = service
        return channel

    def receive_key_value(self):
        return self._service.recv_plist(self)

    def receive_plist(self):
        return self._service.recv_plist(self)[0]

    def receive_message(self):
        return self._service.recv_message(self)[0]

    def send_message(self, selector: str, args: MessageAux = None, expects_reply: bool = True):
        self._service.send_message(self, selector, args, expects_reply=expects_reply)

    @staticmethod
    def _sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        name = "_" + name[1:].replace("_", ":") if name.startswith("_") else name.replace("_", ":")
        return name

    def __getitem__(self, item):
        return partial(self._service.send_message, self, item)

    def __getattr__(self, item):
        return self[self._sanitize_name(item)]


class ChannelFragmenter:
    def __init__(self):
        self._messages = Queue()
        self._packet_data = b""
        self._stream_packet_data = b""

    def get(self):
        return self._messages.get_nowait()

    def add_fragment(self, mheader, chunk):
        self._packet_data += chunk
        if mheader.fragmentId == mheader.fragmentCount - 1:
            # last message
            fake_header = mheader.copy()
            fake_header.fragmentId = 0
            fake_header.fragmentCount = 1
            fake_header.channelCode = abs(mheader.channelCode)
            fake_header.lenght = len(self._packet_data)
            self._messages.put((fake_header, self._packet_data))
            self._packet_data = b""


class RemoteServer(LockdownService):
    """
    Wrapper to Apple's RemoteServer.
    This server exports several ObjC objects allowing calling their respective selectors.
    The `/Developer/Library/PrivateFrameworks/DVTInstrumentsFoundation.framework/DTServiceHub` service reads the
    configuration stored from `[[NSUserDefaults standardUserDefaults] boolForKey:@"DTXConnectionTracer"]`
    If the value is true, then `/tmp/DTServiceHub[PID].DTXConnection.RANDOM.log` is created and can be used to debug the
    transport protocol.

    For example:

    ```
    root@iPhone (/var/root)# tail -f /tmp/DTServiceHub[369].DTXConnection.qNjM2U.log
    170.887982 x4 resuming [c0]: <DTXConnection 0x100d20670 : x4>
    170.889120 x4   sent   [c0]: < DTXMessage 0x100d52b10 : i2.0 c0 dispatch:[_notifyOfPublishedCapabilities:<NSDictionary 0x100d0e1b0 | 92 key/value pairs>] >
    170.889547 x4 received [c0]: < DTXMessage 0x100d0a550 : i1.0 c0 dispatch:[_notifyOfPublishedCapabilities:<NSDictionary 0x100d16a40 | 2 key/value pairs>] >
    170.892101 x4 received [c0]: < DTXMessage 0x100d0a550 : i3.0e c0 dispatch:[_requestChannelWithCode:[1]identifier :"com.apple.instruments.server.services.deviceinfo"] >
    170.892238 x4   sent   [c0]: < DTXMessage 0x100d61830 : i3.1 c0 >
    170.892973 x4 received [c1f]: < DTXMessage 0x100d0a550 : i4.0e c1 dispatch:[runningProcesses] >
    171.204957 x4   sent   [c1f]: < DTXMessage 0x100c557a0 : i4.1 c1 object:(__NSArrayM*)<NSArray 0x100c199d0 | 245 objects> { <NSDictionary 0x100c167c0 | 5 key/value pairs>, <NSDictionary 0x100d17970 | 5 key/value pairs>, <NSDictionary 0x100d17f40 | 5 key/value pairs>, <NSDictionary 0x100d61750 | 5 key/value pairs>, <NSDictionary 0x100c16760 | 5 key/value pairs>, ...  } >
    171.213326 x4 received [c0]: < DTXMessage : kDTXInterruptionMessage >
    171.213424 x4  handler [c0]: < DTXMessage : i1 kDTXInterruptionMessage >
    171.213477 x4 received [c1f]: < DTXMessage : kDTXInterruptionMessage >
    ```

    For editing the configuration we can simply add the respected key into:
    `/var/mobile/Library/Preferences/.GlobalPreferences.plist` and kill `cfprefsd`

    The valid selectors for triggering can be found using the following Frida script the same way Troy Bowman used for
    iterating all classes which implement the protocol `DTXAllowedRPC`:

    ```shell
    frida -U DTServiceHub
    ```

    ```javascript
    for (var name in ObjC.protocols) {
        var protocol = ObjC.protocols[name]
        if ('DTXAllowedRPC' in protocol.protocols) {
            console.log('@protocol', name)
            console.log('  ' + Object.keys(protocol.methods).join('\n  '))
        }
    }
    ```
    """

    BROADCAST_CHANNEL = 0
    INSTRUMENTS_MESSAGE_TYPE = 2

    def __init__(
        self,
        lockdown: LockdownServiceProvider,
        service_name,
        remove_ssl_context: bool = True,
        is_developer_service: bool = True,
    ):
        super().__init__(lockdown, service_name, is_developer_service=is_developer_service)

        if remove_ssl_context and hasattr(self.service.socket, "_sslobj"):
            self.service.socket._sslobj = None

        self.lock = threading.RLock()
        self.supported_identifiers = {}
        self.last_channel_code = 0
        self.cur_message = 0
        self.cur_remote_message = 0
        self.channel_cache = {}
        self.channel_messages = {self.BROADCAST_CHANNEL: ChannelFragmenter()}
        self.broadcast = Channel.create(0, self)

    def _log_dtx_message(self, direction: str, mheader: Container | bytes, payload: bytes) -> None:
        if not self.logger.isEnabledFor(logging.DEBUG):
            return

        if isinstance(mheader, bytes):
            mheader = dtx_message_header_struct.parse(mheader)

        s = io.BytesIO(payload)
        pheader = dtx_message_payload_header_struct.parse_stream(s)
        compression = pheader.flags == 0x0707
        if compression:
            raise NotImplementedError("Compressed")

        aux = message_aux_t_struct.parse_stream(s).aux if pheader.auxiliaryLength else None
        obj_size = pheader.totalLength - pheader.auxiliaryLength
        data = s.read(obj_size) if obj_size else None

        with suppress(Exception):
            if pheader.auxiliaryLength:
                aux = [c.value for c in aux]
        with suppress(Exception):
            if obj_size:
                data = archiver.unarchive(data)

        e = "e" if mheader.expectsReply else ""
        self.logger.debug(
            "x%d %-8s [c?]: <DTXMessage: i%d.%d%s c%d t:%d payload:%s aux:%s>",
            self.service.socket.fileno(),
            direction,
            mheader.identifier,
            mheader.conversationIndex,
            e,
            mheader.channelCode,
            pheader.flags,
            data,
            aux,
        )

    def shell(self):
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.Terminal256Formatter(style="native")),
            user_ns={
                "developer": self,
                "broadcast": self.broadcast,
                "MessageAux": MessageAux,
            },
        )

    def perform_handshake(self):
        args = MessageAux()
        args.append_obj({"com.apple.private.DTXBlockCompression": 0, "com.apple.private.DTXConnection": 1})

        with self.lock:
            self.send_message(0, "_notifyOfPublishedCapabilities:", args, expects_reply=False)
            ret, aux = self.recv_plist()
            if ret != "_notifyOfPublishedCapabilities:":
                raise ValueError("Invalid answer")
            if not len(aux[0]):
                raise ValueError("Invalid answer")
            self.supported_identifiers = aux[0].value

    def make_channel(self, identifier) -> Channel:
        # NOTE: There is also identifier not in self.supported_identifiers
        # assert identifier in self.supported_identifiers
        with self.lock:
            if identifier in self.channel_cache:
                return self.channel_cache[identifier]

            self.last_channel_code += 1
            code = self.last_channel_code
            args = MessageAux().append_int(code).append_obj(identifier)
            self.send_message(0, "_requestChannelWithCode:identifier:", args)
            ret, _aux = self.recv_plist()
            assert ret is None
            channel = Channel.create(code, self)
            self.channel_cache[identifier] = channel
            self.channel_messages[code] = ChannelFragmenter()
            return channel

    def serve_channel(self, identifier: str, channel: Optional[Channel] = None) -> Channel:
        """acknoledge the remote part that they will find the expected identifier on the given channel"""
        with self.lock:
            msg = self.recv_plist(self.BROADCAST_CHANNEL)
            assert msg[0] == "_requestChannelWithCode:identifier:", (
                f"expected a request for a reverse channel, got: {msg}"
            )
            code = msg[1][0].value
            assert channel is None or channel == code, (
                f"expected a request for a reverse channel with channelCode:{channel}, got:{msg[1][0]}"
            )
            identifier = msg[1][1].value
            assert identifier == msg[1][1].value, (
                f"expected a request for a reverse channel with identifier:{identifier}, got:{msg[1][1].value}"
            )
            if channel is None:
                channel = Channel.create(code, self)
            if code not in self.channel_messages:
                self.channel_messages[code] = ChannelFragmenter()
            self.send_reply_ack(self.BROADCAST_CHANNEL, self.cur_remote_message, None, None)
            return channel

    def send_message(
        self, channel: int, selector: Optional[str] = None, args: MessageAux = None, expects_reply: bool = True
    ):
        aux = bytes(args) if args is not None else b""
        sel = archiver.archive(selector) if selector is not None else b""
        flags = self.INSTRUMENTS_MESSAGE_TYPE
        pheader = dtx_message_payload_header_struct.build({
            "flags": flags,
            "auxiliaryLength": len(aux),
            "totalLength": len(aux) + len(sel),
        })

        with self.lock:
            self.cur_message += 1
            mheader = dtx_message_header_struct.build({
                "cb": dtx_message_header_struct.sizeof(),
                "fragmentId": 0,
                "fragmentCount": 1,
                "length": dtx_message_payload_header_struct.sizeof() + len(aux) + len(sel),
                "identifier": self.cur_message,
                "conversationIndex": 0,
                "channelCode": channel,
                "expectsReply": int(expects_reply),
            })
            msg = mheader + pheader + aux + sel
            self.service.sendall(msg)

            self._log_dtx_message("sent", mheader, pheader + aux + sel)

    def recv_plist(self, channel: int = BROADCAST_CHANNEL):
        data, aux = self.recv_message(channel)
        if data is not None:
            try:
                data = archiver.unarchive(data)
            except archiver.MissingClassMapping:
                pprint(plistlib.loads(data))
                raise
            except plistlib.InvalidFileException:
                self.logger.warning(f"got an invalid plist: {data[:40]}")
        return data, aux

    def recv_message(self, channel: int = BROADCAST_CHANNEL):
        packet_stream = self._recv_packet_fragments(channel)
        pheader = dtx_message_payload_header_struct.parse_stream(packet_stream)

        compression = (pheader.flags & 0xFF000) >> 12
        if compression:
            raise NotImplementedError("Compressed")

        aux = message_aux_t_struct.parse_stream(packet_stream).aux if pheader.auxiliaryLength else None
        obj_size = pheader.totalLength - pheader.auxiliaryLength
        data = packet_stream.read(obj_size) if obj_size else None
        return data, aux

    # TODO: rewrite the RemoteServer class (possibly even ServiceConnection) to continue consume messages
    #     : and dispatch them asyncronously to the opened channels. Each channel is connected to a service and
    #     : the DVT client ( us ) shall be able to provide some services as well. ( f.i. dtxproxy:XCTestDriverInterface:XCTestManager_IDEInterface ).
    #     : It's important to note that the current implementation suppose that messages are sent and consumed in order and sequentially,
    #     : as soon as you start multiple threads on the same connection, everything breaks down.
    #     : This will require an extensive refactoring as the message identifier and conversation index must be kept until a reply
    #     : is ready to send with conversationIndex+1 and the same message identifier.
    #     : Draft: ReaderStream.recv_one -> srv = services_by_chan[chan] & send_reply(msg.id, srv.__call_method__(payload, aux))
    #     : Draft: ReaderStream.recv_one -> replies[msg.id].put(msg) & wait_reply_for_msg(msg.id) -> msg
    # FIXME: messages are independent of channels, they have an unique identifier that is reused only for replies (conversationIndex > 0).
    #      : therefore the defragmenter shall work at the connection-level and store fragments by using this as hash: (msg.identifier, msg.conversationIndex).
    #      : please note that the 2 ends of the connection have their own message identifier counter.

    def _send_reply(
        self,
        channel: int,
        msg_id: int,
        msg_type: int,
        payload: Optional[object] = None,
        aux: Optional[MessageAux] = None,
    ):
        payload_bytes = archiver.archive(payload) if payload is not None else b""
        aux_bytes = archiver.archive(aux) if aux is not None else b""
        pheader = dtx_message_payload_header_struct.build({
            "flags": msg_type,
            "auxiliaryLength": len(aux_bytes),
            "totalLength": len(payload_bytes) + len(aux_bytes),
        })
        mheader = dtx_message_header_struct.build({
            "cb": dtx_message_header_struct.sizeof(),
            "fragmentId": 0,
            "fragmentCount": 1,
            "length": dtx_message_payload_header_struct.sizeof() + len(payload_bytes) + len(aux_bytes),
            "identifier": msg_id,
            "conversationIndex": 1,
            "channelCode": channel,
            "expectsReply": (0),
        })
        msg = mheader + pheader + aux_bytes + payload_bytes
        with self.lock:
            self.service.sendall(msg)
            self._log_dtx_message("sent", mheader, pheader + aux_bytes + payload_bytes)

    def send_reply(self, channel: int, msg_id: int, payload: Optional[object] = None, aux: Optional[MessageAux] = None):
        msg_type = 0x3  # ResponseWithReturnValueInPayload
        self._send_reply(channel, msg_id, msg_type, payload, aux)

    def send_reply_error(
        self, channel: int, msg_id: int, payload: Optional[object] = None, aux: Optional[MessageAux] = None
    ):
        msg_type = 0x4  # DtxTypeError
        self._send_reply(channel, msg_id, msg_type, payload, aux)

    def send_reply_ack(
        self, channel: int, msg_id: int, payload: Optional[object] = None, aux: Optional[MessageAux] = None
    ):
        msg_type = 0x0  # Ack
        self._send_reply(channel, msg_id, msg_type, payload, aux)

    def _recv_packet_fragments(self, channel: int = BROADCAST_CHANNEL):
        while True:
            try:
                # if we already have a message for this channel, just return it
                with self.lock:
                    # TODO: pthread_cond_wait with a background thread consuming messages
                    mheader, message = self.channel_messages[channel].get()
                    self.cur_remote_message = mheader.identifier
                self._log_dtx_message("received", mheader, message)
                return io.BytesIO(message)
            except Empty:
                # if no message exists for the given channel code, just keep waiting and receive new messages
                # until the waited message queue has at least one message
                with self.lock:
                    data = self.service.recvall(dtx_message_header_struct.sizeof())
                    mheader = dtx_message_header_struct.parse(data)

                    # treat both as the negative and positive representation of the channel code in the response
                    # the same when performing fragmentation
                    received_channel_code = abs(mheader.channelCode)

                    if received_channel_code not in self.channel_messages:
                        self.channel_messages[received_channel_code] = ChannelFragmenter()

                    if not mheader.conversationIndex and mheader.identifier > self.cur_message:
                        self.cur_message = mheader.identifier

                    if mheader.fragmentCount > 1 and mheader.fragmentId == 0:
                        # when reading multiple message fragments, the first fragment contains only a message header
                        continue

                    self.channel_messages[received_channel_code].add_fragment(
                        mheader, self.service.recvall(mheader.length)
                    )

    def __enter__(self):
        self.perform_handshake()
        return self

    def close(self):
        aux = MessageAux()
        codes = [code for code in self.channel_messages if code > 0]
        if codes:
            for code in codes:
                aux.append_int(code)

            with contextlib.suppress(OSError):
                # ignore: OSError: [Errno 9] Bad file descriptor
                self.send_message(self.BROADCAST_CHANNEL, "_channelCanceled:", aux, expects_reply=False)
        super().close()


class Tap:
    def __init__(self, dvt, channel_name: str, config: dict):
        self._dvt = dvt
        self._channel_name = channel_name
        self._config = config
        self.channel = None

    def __enter__(self):
        self.channel = self._dvt.make_channel(self._channel_name)
        self.channel.setConfig_(MessageAux().append_obj(self._config), expects_reply=False)
        self.channel.start(expects_reply=False)

        # first message is just kind of an ack
        self.channel.receive_plist()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.channel.clear(expects_reply=False)

    def __iter__(self):
        while True:
            yield from self.channel.receive_plist()



import asyncio
import contextlib
import logging
import plistlib
import socket
import ssl
import struct
import time
import xml
from enum import Enum
from typing import Any, Optional, Union

import IPython
from pygments import formatters, highlight, lexers

from pymobiledevice3.exceptions import (
    ConnectionTerminatedError,
    DeviceNotFoundError,
    NoDeviceConnectedError,
    PyMobileDevice3Exception,
)
from pymobiledevice3.osu.os_utils import get_os_utils
from pymobiledevice3.usbmux import MuxDevice, select_device

DEFAULT_AFTER_IDLE_SEC = 3
DEFAULT_INTERVAL_SEC = 3
DEFAULT_MAX_FAILS = 3
DEFAULT_TIMEOUT = 1
OSUTIL = get_os_utils()
SHELL_USAGE = """
# This shell allows you to communicate directly with every service layer behind the lockdownd daemon.

# For example, you can do the following:
client.send_plist({"Command": "DoSomething"})

# and view the reply
print(client.recv_plist())

# or just send raw message
client.send(b"hello")

# and view the result
print(client.recvall(20))
"""


def build_plist(d: dict, endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> bytes:
    """
    Convert a dictionary to a plist-formatted byte string prefixed with a length field.

    :param d: The dictionary to convert.
    :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
    :param fmt: The plist format (e.g., plistlib.FMT_XML).
    :return: The plist-formatted byte string.
    """
    payload = plistlib.dumps(d, fmt=fmt)
    message = struct.pack(endianity + "L", len(payload))
    return message + payload


def parse_plist(payload: bytes) -> dict:
    """
    Parse a plist-formatted byte string into a dictionary.

    :param payload: The plist-formatted byte string to parse.
    :return: The parsed dictionary.
    :raises PyMobileDevice3Exception: If the payload is invalid.
    :retries with a filtered payload of only valid XML characters if plistlib compains about "not well-formed (invalid token)"
    """
    try:
        return plistlib.loads(payload)
    except xml.parsers.expat.ExpatError:
        payload = bytes([b for b in payload if b >= 0x20 or b in (0x09, 0x0A, 0x0D)])
        try:
            return plistlib.loads(payload)
        except plistlib.InvalidFileException as e:
            raise PyMobileDevice3Exception(f"parse_plist invalid data: {payload[:100].hex()}") from e
    except plistlib.InvalidFileException as e:
        raise PyMobileDevice3Exception(f"parse_plist invalid data: {payload[:100].hex()}") from e


class ServiceConnection:
    """wrapper for tcp-relay connections"""

    def __init__(self, sock: socket.socket, mux_device: MuxDevice = None):
        """
        Initialize a ServiceConnection object.

        :param sock: The socket to use for the connection.
        :param mux_device: The MuxDevice associated with the connection (optional).
        """
        self.logger = logging.getLogger(__name__)
        self.socket = sock
        self._offset = 0

        # usbmux connections contain additional information associated with the current connection
        self.mux_device = mux_device

        self.reader = None  # type: Optional[asyncio.StreamReader]
        self.writer = None  # type: Optional[asyncio.StreamWriter]

        # SSL/TLS version to be used for connecting to device
        # TLS v1.2 is supported since iOS 5
        self.min_ssl_proto = ssl.TLSVersion.TLSv1_2
        self.max_ssl_proto = ssl.TLSVersion.TLSv1_3

    @staticmethod
    def create_using_tcp(
        hostname: str, port: int, keep_alive: bool = True, create_connection_timeout: int = DEFAULT_TIMEOUT
    ) -> "ServiceConnection":
        """
        Create a ServiceConnection using a TCP connection.

        :param hostname: The hostname of the server to connect to.
        :param port: The port to connect to.
        :param keep_alive: Whether to enable TCP keep-alive.
        :param create_connection_timeout: The timeout for creating the connection.
        :return: A ServiceConnection object.
        """
        sock = socket.create_connection((hostname, port), timeout=create_connection_timeout)
        sock.settimeout(None)
        if keep_alive:
            OSUTIL.set_keepalive(sock)
        return ServiceConnection(sock)

    @staticmethod
    def create_using_usbmux(
        udid: Optional[str], port: int, connection_type: Optional[str] = None, usbmux_address: Optional[str] = None
    ) -> "ServiceConnection":
        """
        Create a ServiceConnection using a USBMux connection.

        :param udid: The UDID of the target device.
        :param port: The port to connect to.
        :param connection_type: The type of connection to use.
        :param usbmux_address: The address of the usbmuxd socket.
        :return: A ServiceConnection object.
        :raises DeviceNotFoundError: If the device with the specified UDID is not found.
        :raises NoDeviceConnectedError: If no device is connected.
        """
        target_device = select_device(udid, connection_type=connection_type, usbmux_address=usbmux_address)
        if target_device is None:
            if udid:
                raise DeviceNotFoundError(udid)
            raise NoDeviceConnectedError()
        sock = target_device.connect(port, usbmux_address=usbmux_address)
        return ServiceConnection(sock, mux_device=target_device)

    def setblocking(self, blocking: bool) -> None:
        """
        Set the blocking mode of the socket.

        :param blocking: If True, set the socket to blocking mode; otherwise, set it to non-blocking mode.
        """
        self.socket.setblocking(blocking)

    def close(self) -> None:
        """Close the connection."""
        self.socket.close()

    async def aio_close(self) -> None:
        """Asynchronously close the connection."""
        if self.writer is None:
            return
        self.writer.close()
        with contextlib.suppress(ssl.SSLError):
            await self.writer.wait_closed()
        self.writer = None
        self.reader = None

    def recv(self, length: int = 4096) -> bytes:
        """
        Receive data from the socket.

        :param length: The maximum amount of data to receive.
        :return: The received data.
        """
        try:
            return self.socket.recv(length)
        except (ssl.SSLError, BrokenPipeError) as e:
            raise ConnectionAbortedError() from e

    def sendall(self, data: bytes) -> None:
        """
        Send data to the socket.

        :param data: The data to send.
        :raises ConnectionTerminatedError: If the connection is terminated abruptly.
        """
        try:
            self.socket.sendall(data)
        except ssl.SSLEOFError as e:
            raise ConnectionTerminatedError from e

    def send_recv_plist(self, data: Union[dict, list], endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> Any:
        """
        Send a plist to the socket and receive a plist response.

        :param data: The dictionary to send as a plist.
        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :param fmt: The plist format (e.g., plistlib.FMT_XML).
        :return: The received plist as a dictionary.
        """
        self.send_plist(data, endianity=endianity, fmt=fmt)
        return self.recv_plist(endianity=endianity)

    async def aio_send_recv_plist(self, data: dict, endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> Any:
        """
        Asynchronously send a plist to the socket and receive a plist response.

        :param data: The dictionary to send as a plist.
        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :param fmt: The plist format (e.g., plistlib.FMT_XML).
        :return: The received plist as a dictionary.
        """
        await self.aio_send_plist(data, endianity=endianity, fmt=fmt)
        return await self.aio_recv_plist(endianity=endianity)

    def recvall(self, size: int) -> bytes:
        """
        Receive all data of a specified size from the socket.

        :param size: The amount of data to receive.
        :return: The received data.
        :raises ConnectionAbortedError: If the connection is aborted.
        """
        data = b""
        while len(data) < size:
            chunk = self.recv(size - len(data))
            if chunk is None or len(chunk) == 0:
                raise ConnectionAbortedError()
            data += chunk
        return data

    def recv_prefixed(self, endianity: str = ">") -> bytes:
        """
        Receive a data block prefixed with a length field.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received data block.
        """
        size = self.recvall(4)
        if not size or len(size) != 4:
            return b""
        size = struct.unpack(endianity + "L", size)[0]
        while True:
            try:
                return self.recvall(size)
            except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                # Allow ssl to do stuff
                time.sleep(0)

    async def aio_recvall(self, size: int) -> bytes:
        """
        Asynchronously receive data of a specified size from the socket.

        :param size: The amount of data to receive.
        :return: The received data.
        """
        return await self.reader.readexactly(size)

    async def aio_recv_prefixed(self, endianity: str = ">") -> bytes:
        """
        Asynchronously receive a data block prefixed with a length field.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received data block.
        """
        size = await self.aio_recvall(4)
        size = struct.unpack(endianity + "L", size)[0]
        return await self.aio_recvall(size)

    def send_prefixed(self, data: bytes) -> None:
        """
        Send a data block prefixed with a length field.

        :param data: The data to send.
        """
        if isinstance(data, str):
            data = data.encode()
        hdr = struct.pack(">L", len(data))
        msg = b"".join([hdr, data])
        return self.sendall(msg)

    def recv_plist(self, endianity: str = ">") -> Union[dict, list]:
        """
        Receive a plist from the socket and parse it into a native type.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received plist as a native type.
        """
        return parse_plist(self.recv_prefixed(endianity=endianity))

    async def aio_recv_plist(self, endianity: str = ">") -> dict:
        """
        Asynchronously receive a plist from the socket and parse it into a native type.

        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :return: The received plist as a native type.
        """
        return parse_plist(await self.aio_recv_prefixed(endianity))

    def send_plist(self, d: Union[dict, list], endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> None:
        """
        Send a native type as a plist to the socket.

        :param d: The native type to send.
        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :param fmt: The plist format (e.g., plistlib.FMT_XML).
        """
        return self.sendall(build_plist(d, endianity, fmt))

    async def aio_sendall(self, payload: bytes) -> None:
        """
        Asynchronously send data to the socket.

        :param payload: The data to send.
        """
        self.writer.write(payload)
        await self.writer.drain()

    async def aio_send_plist(self, d: Union[dict, list], endianity: str = ">", fmt: Enum = plistlib.FMT_XML) -> None:
        """
        Asynchronously send a dictionary as a plist to the socket.

        :param d: The dictionary to send.
        :param endianity: The byte order ('>' for big-endian, '<' for little-endian).
        :param fmt: The plist format (e.g., plistlib.FMT_XML).
        """
        await self.aio_sendall(build_plist(d, endianity, fmt))

    def create_ssl_context(self, certfile: str, keyfile: Optional[str] = None) -> ssl.SSLContext:
        """
        Create an SSL context for a secure connection.

        :param certfile: The path to the certificate file.
        :param keyfile: The path to the key file (optional).
        :return: An SSL context object.
        """
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.minimum_version = self.min_ssl_proto
        context.maximum_version = self.max_ssl_proto
        if ssl.OPENSSL_VERSION.lower().startswith("openssl"):
            context.set_ciphers("ALL:!aNULL:!eNULL:@SECLEVEL=0")
        else:
            context.set_ciphers("ALL:!aNULL:!eNULL")
        context.options |= 0x4  # OPENSSL OP_LEGACY_SERVER_CONNECT (required for legacy iOS devices)
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        context.load_cert_chain(certfile, keyfile)
        return context

    def ssl_start(self, certfile: str, keyfile: Optional[str] = None) -> None:
        """
        Start an SSL connection.

        :param certfile: The path to the certificate file.
        :param keyfile: The path to the key file (optional).
        """
        try:
            self.socket = self.create_ssl_context(certfile, keyfile=keyfile).wrap_socket(self.socket)
        except OSError as e:
            raise ConnectionAbortedError() from e

    async def aio_ssl_start(self, certfile: str, keyfile: Optional[str] = None) -> None:
        """
        Asynchronously start an SSL connection.

        :param certfile: The path to the certificate file.
        :param keyfile: The path to the key file (optional).
        """
        self.reader, self.writer = await asyncio.open_connection(
            sock=self.socket, ssl=self.create_ssl_context(certfile, keyfile=keyfile), server_hostname=""
        )

    async def aio_start(self) -> None:
        """Asynchronously start a connection."""
        self.reader, self.writer = await asyncio.open_connection(sock=self.socket)

    def shell(self) -> None:
        """Start an interactive shell."""
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.Terminal256Formatter(style="native")),
            user_ns={
                "client": self,
            },
        )

    def read(self, size: int) -> bytes:
        """
        Read data from the socket.

        :param size: The amount of data to read.
        :return: The read data.
        """
        result = self.recvall(size)
        self._offset += size
        return result

    def write(self, data: bytes) -> None:
        """
        Write data to the socket.

        :param data: The data to write.
        """
        self.sendall(data)
        self._offset += len(data)

    def tell(self) -> int:
        """
        Get the current offset.

        :return: The current offset.
        """
        return self._offset



import importlib.resources
import json
import logging
from itertools import islice
from pathlib import Path
from typing import Annotated, Optional

import typer
from pykdebugparser.pykdebugparser import PyKdebugParser
from typer_injector import InjectingTyper

import pymobiledevice3.resources
from pymobiledevice3.cli.cli_common import (
    BASED_INT,
    ServiceProviderDep,
    default_json_encoder,
    print_json,
    user_requested_colored_output,
)
from pymobiledevice3.exceptions import ExtractingStackshotError
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.core_profile_session_tap import CoreProfileSessionTap
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo

logger = logging.getLogger(__name__)

BSC_SUBCLASS = 0x40C
BSC_CLASS = 0x4
VFS_AND_TRACES_SET = {0x03010000, 0x07FF0000}


cli = InjectingTyper(
    name="core-profile-session",
    help="Access tailspin features",
    no_args_is_help=True,
)


BSCFilter = Annotated[
    bool,
    typer.Option(help="Whether to print BSC events or not."),
]
ClassFilter = Annotated[
    list[int],
    typer.Option(
        "--class-filters",
        "-cf",
        click_type=BASED_INT,
        default_factory=list,
        show_default=False,
        help="Events class filter. Omit for all. Can be specified multiple times.",
    ),
]
SubclassFilter = Annotated[
    list[int],
    typer.Option(
        "--subclass-filters",
        "-sf",
        click_type=BASED_INT,
        default_factory=list,
        show_default=False,
        help="Events subclass filter. Omit for all. Can be specified multiple times.",
    ),
]
Count = Annotated[
    Optional[int],
    typer.Option(
        "--count",
        "-c",
        help="Number of events to print. Omit to endless sniff.",
    ),
]
ThreadID = Annotated[
    Optional[int],
    typer.Option(help="Thread ID to filter. Omit for all."),
]
ShowThreadID = Annotated[
    bool,
    typer.Option(help="Whether to print thread ID or not."),
]


def parse_filters(subclasses: list[int], classes: list[int]) -> Optional[set[int]]:
    if not subclasses and not classes:
        return None
    parsed: set[int] = set()
    for subclass in subclasses:
        if subclass == BSC_SUBCLASS:
            parsed |= VFS_AND_TRACES_SET
        parsed.add(subclass << 16)
    for class_ in classes:
        if class_ == BSC_CLASS:
            parsed |= VFS_AND_TRACES_SET
        parsed.add((class_ << 24) | 0x00FF0000)
    return parsed


@cli.command("live")
def live_profile_session(
    service_provider: ServiceProviderDep,
    *,
    count: Count = -1,
    bsc: BSCFilter = False,
    class_filters: ClassFilter,
    subclass_filters: SubclassFilter,
    tid: ThreadID = None,
    timestamp: Annotated[
        bool,
        typer.Option(help="Whether to print timestamp or not."),
    ] = True,
    event_name: Annotated[
        bool,
        typer.Option(help="Whether to print event name or not."),
    ] = True,
    func_qual: Annotated[
        bool,
        typer.Option(help="Whether to print function qualifier or not."),
    ] = True,
    show_tid: ShowThreadID = True,
    process_name: Annotated[
        bool,
        typer.Option(help="Whether to print process name or not."),
    ] = True,
    args: Annotated[
        bool,
        typer.Option(help="Whether to print event arguments or not."),
    ] = True,
) -> None:
    """Print kevents received from the device in real time."""
    parser = PyKdebugParser()
    class_filters = class_filters
    subclass_filters = subclass_filters
    parser.filter_class = class_filters
    if bsc:
        subclass_filters.append(BSC_SUBCLASS)
    parser.filter_subclass = subclass_filters
    filters = parse_filters(subclass_filters, class_filters)
    parser.filter_tid = tid
    parser.show_timestamp = timestamp
    parser.show_name = event_name
    parser.show_func_qual = func_qual
    parser.show_tid = show_tid
    parser.show_process = process_name
    parser.show_args = args
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        trace_codes_map = DeviceInfo(dvt).trace_codes()
        time_config = CoreProfileSessionTap.get_time_config(dvt)
        parser.numer = time_config["numer"]
        parser.denom = time_config["denom"]
        parser.mach_absolute_time = time_config["mach_absolute_time"]
        parser.usecs_since_epoch = time_config["usecs_since_epoch"]
        parser.timezone = time_config["timezone"]
        with CoreProfileSessionTap(dvt, time_config, filters) as tap:
            for i, event in enumerate(parser.formatted_kevents(tap.get_kdbuf_stream(), trace_codes_map)):
                print(event)
                if i == count:
                    break


@cli.command("save")
def save_profile_session(
    service_provider: ServiceProviderDep,
    out: Path,
    *,
    bsc: BSCFilter = False,
    class_filters: ClassFilter,
    subclass_filters: SubclassFilter,
) -> None:
    """Dump core profiling information."""
    if bsc:
        subclass_filters.append(BSC_SUBCLASS)
    filters = parse_filters(subclass_filters, class_filters)
    with (
        DvtSecureSocketProxyService(lockdown=service_provider) as dvt,
        CoreProfileSessionTap(dvt, {}, filters) as tap,
        out.open("wb") as out_file,
    ):
        tap.dump(out_file)


@cli.command("stackshot")
def stackshot(
    service_provider: ServiceProviderDep,
    out: Annotated[Optional[Path], typer.Option()] = None,
) -> None:
    """Dump stackshot information."""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt, CoreProfileSessionTap(dvt, {}) as tap:
        try:
            data = tap.get_stackshot()
        except ExtractingStackshotError:
            logger.exception("Extracting stackshot failed")
            return

        if out is not None:
            out.write_text(json.dumps(data, indent=4, default=default_json_encoder))
        else:
            print_json(data)


@cli.command("parse-live")
def parse_live_profile_session(
    service_provider: ServiceProviderDep,
    *,
    count: Count = None,
    bsc: BSCFilter = False,
    class_filters: ClassFilter,
    subclass_filters: SubclassFilter,
    tid: ThreadID = None,
    show_tid: ShowThreadID = False,
    process: Annotated[
        Optional[str],
        typer.Option(help="Process ID / name to filter. Omit for all."),
    ] = None,
) -> None:
    """Print traces (syscalls, thread events, etc.) received from the device in real time."""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        print("Receiving time information")
        time_config = CoreProfileSessionTap.get_time_config(dvt)
        parser = PyKdebugParser()
        parser.filter_class = list(class_filters)
        if bsc:
            subclass_filters.append(BSC_SUBCLASS)
        parser.filter_subclass = subclass_filters
        filters = parse_filters(subclass_filters, class_filters)
        parser.numer = time_config["numer"]
        parser.denom = time_config["denom"]
        parser.mach_absolute_time = time_config["mach_absolute_time"]
        parser.usecs_since_epoch = time_config["usecs_since_epoch"]
        parser.timezone = time_config["timezone"]
        parser.filter_tid = tid
        parser.filter_process = process
        parser.show_tid = show_tid
        parser.color = user_requested_colored_output()

        with CoreProfileSessionTap(dvt, time_config, filters) as tap:
            if show_tid:
                print("{:^32}|{:^11}|{:^33}|   Event".format("Time", "Thread", "Process"))
            else:
                print("{:^32}|{:^33}|   Event".format("Time", "Process"))

            for trace in islice(parser.formatted_traces(tap.get_kdbuf_stream()), count):
                print(trace, flush=True)


def get_image_name(dsc_uuid_map, image_uuid, current_dsc_map):
    if not current_dsc_map:
        for dsc_mapping in dsc_uuid_map.values():
            if image_uuid in dsc_mapping:
                current_dsc_map.update(dsc_mapping)

    return current_dsc_map.get(image_uuid, image_uuid)


def format_callstack(callstack: str, dsc_uuid_map, current_dsc_map) -> str:
    lines = callstack.splitlines()
    for i, line in enumerate(lines[1:]):
        if ":" in line:
            uuid = line.split(":")[0].strip()
            lines[i + 1] = line.replace(uuid, get_image_name(dsc_uuid_map, uuid, current_dsc_map))
    return "\n".join(lines)


@cli.command("callstacks-live")
def callstacks_live_profile_session(
    service_provider: ServiceProviderDep,
    count: Count = -1,
    process: Annotated[
        Optional[str],
        typer.Option(help="Process ID / name to filter. Omit for all."),
    ] = None,
    tid: ThreadID = None,
    show_tid: ShowThreadID = False,
) -> None:
    """Print callstacks received from the device in real time."""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        print("Receiving time information")
        time_config = CoreProfileSessionTap.get_time_config(dvt)
        parser = PyKdebugParser()
        parser.numer = time_config["numer"]
        parser.denom = time_config["denom"]
        parser.mach_absolute_time = time_config["mach_absolute_time"]
        parser.usecs_since_epoch = time_config["usecs_since_epoch"]
        parser.timezone = time_config["timezone"]
        parser.filter_tid = tid
        parser.filter_process = process
        parser.color = user_requested_colored_output()
        parser.show_tid = show_tid

        with importlib.resources.open_text(pymobiledevice3.resources, "dsc_uuid_map.json") as fd:
            dsc_uuid_map = json.load(fd)

        current_dsc_map = {}
        with CoreProfileSessionTap(dvt, time_config) as tap:
            for callstack in islice(parser.formatted_callstacks(tap.get_kdbuf_stream()), count):
                print(format_callstack(callstack, dsc_uuid_map, current_dsc_map))



import plistlib

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.lockdown_service import LockdownService

CHUNK_SIZE = 200


class DebugServerAppList(LockdownService):
    SERVICE_NAME = "com.apple.debugserver.DVTSecureSocketProxy.applist"

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def get(self) -> dict:
        buf = b""
        while b"</plist>" not in buf:
            buf += self.service.recv(CHUNK_SIZE)

        return plistlib.loads(buf)



from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.core_profile_session_tap import (
    CoreProfileSessionTap,
)


def test_stackshot(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test getting stackshot.
    """
    with CoreProfileSessionTap(dvt, CoreProfileSessionTap.get_time_config(dvt)) as tap:
        data = tap.get_stackshot()

    assert "Darwin Kernel" in data["osversion"]
    # Constant kernel task data.
    assert data["task_snapshots"][0]["task_snapshot"]["ts_pid"] == 0
    assert data["task_snapshots"][0]["task_snapshot"]["ts_p_comm"] == "kernel_task"



import contextlib
import json
import logging
import time
from collections import namedtuple
from datetime import datetime, timezone
from pathlib import Path
from typing import Annotated, Optional

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, default_json_encoder, print_json
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.sysmontap import Sysmontap

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="process",
    help="Process monitor options.",
    no_args_is_help=True,
)


@cli.command("monitor")
def sysmon_process_monitor(service_provider: ServiceProviderDep, threshold: float) -> None:
    """monitor all most consuming processes by given cpuUsage threshold."""

    Process = namedtuple("process", "pid name cpuUsage physFootprint")

    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt, Sysmontap(dvt) as sysmon:
        for process_snapshot in sysmon.iter_processes():
            entries = []
            for process in process_snapshot:
                if (process["cpuUsage"] is not None) and (process["cpuUsage"] >= threshold):
                    entries.append(
                        Process(
                            pid=process["pid"],
                            name=process["name"],
                            cpuUsage=process["cpuUsage"],
                            physFootprint=process["physFootprint"],
                        )
                    )

            logger.info(entries)


@cli.command("single")
def sysmon_process_single(
    service_provider: ServiceProviderDep,
    attributes: Annotated[
        Optional[list[str]],
        typer.Option(
            "--attributes",
            "-a",
            help="filter processes by given attribute value given as key=value. Can be specified multiple times.",
        ),
    ] = None,
) -> None:
    """show a single snapshot of currently running processes."""

    count = 0

    result = []
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        device_info = DeviceInfo(dvt)

        with Sysmontap(dvt) as sysmon:
            for process_snapshot in sysmon.iter_processes():
                count += 1

                if count < 2:
                    # first sample doesn't contain an initialized value for cpuUsage
                    continue

                for process in process_snapshot:
                    skip = False
                    if attributes is not None:
                        for filter_attr in attributes:
                            filter_attr, filter_value = filter_attr.split("=", 1)
                            if str(process[filter_attr]) != filter_value:
                                skip = True
                                break

                    if skip:
                        continue

                    # adding "artificially" the execName field
                    process["execName"] = device_info.execname_for_pid(process["pid"])
                    result.append(process)

                # exit after single snapshot
                break

    print_json(result)


@cli.command("monitor-single")
def sysmon_process_monitor_single(
    service_provider: ServiceProviderDep,
    attributes: Annotated[
        Optional[list[str]],
        typer.Option(
            "--attributes",
            "-a",
            help="filter processes by attribute (key=value). Multiple filters on same attribute use OR logic, different attributes use AND.",
        ),
    ] = None,
    output: Annotated[
        Optional[Path],
        typer.Option(
            "--output",
            "-o",
            help="output file path for JSONL format (optional, defaults to stdout)",
        ),
    ] = None,
    interval: Annotated[
        Optional[int],
        typer.Option(
            "--interval",
            "-i",
            help="minimum interval in milliseconds between outputs (optional)",
        ),
    ] = None,
    duration: Annotated[
        Optional[int],
        typer.Option(
            "--duration",
            "-d",
            help="maximum duration in milliseconds to run monitoring (optional)",
        ),
    ] = None,
) -> None:
    """Continuously monitor a single process with comprehensive metrics."""
    count = 0
    start_time = None

    # Parse attributes into grouped filters: same attribute uses OR, different attributes use AND
    parsed_filters: dict[str, list[str]] = {}
    if attributes:
        for raw in attributes:
            key, value = raw.split("=", 1)
            parsed_filters.setdefault(key, []).append(value)

    def matches_filters(proc: dict) -> bool:
        """Check if process matches all filter criteria."""
        if not parsed_filters:
            return True
        return all(str(proc.get(key)) in values for key, values in parsed_filters.items())

    with contextlib.ExitStack() as stack:
        output_file = stack.enter_context(open(output, "w")) if output else None

        dvt = stack.enter_context(DvtSecureSocketProxyService(lockdown=service_provider))
        sysmon = stack.enter_context(Sysmontap(dvt))

        for process_snapshot in sysmon.iter_processes():
            count += 1

            if count < 2:
                continue

            if start_time is None:
                start_time = time.time()

            if duration is not None:
                elapsed_ms = (time.time() - start_time) * 1000
                if elapsed_ms >= duration:
                    break

            for process in process_snapshot:
                if not matches_filters(process):
                    continue

                process["timestamp"] = datetime.now(timezone.utc).isoformat()

                if output_file:
                    json_output = json.dumps(process, default=default_json_encoder)
                    output_file.write(json_output + "\n")
                    output_file.flush()
                else:
                    print_json(process)

            if interval:
                time.sleep(interval / 1000.0)
